"""Utilities functions for query  
"""
import json
import os, sys

try:
    from urllib.error import HTTPError
    from urllib.parse import quote, urlencode
    from urllib.request import urlopen
except ImportError:
    from urllib import urlencode
    from urllib2 import quote, urlopen, HTTPError


def _parse_prop(search, proplist):
    """Extract property value from record using the given urn search filter."""
    props = [
        i for i in proplist
        if all(item in i["urn"].items() for item in search.items())
    ]
    if len(props) > 0:
        return props[0]["value"][list(props[0]["value"].keys())[0]]


def request(
    identifier,
    namespace="cid",
    domain="compound",
    operation=None,
    output="JSON",
    searchtype=None,
):
    """
    copied from https://github.com/mcs07/PubChemPy/blob/e3c4f4a9b6120433e5cc3383464c7a79e9b2b86e/pubchempy.py#L238
    Construct API request from parameters and return the response.
    Full specification at http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html
    """
    API_BASE = "https://pubchem.ncbi.nlm.nih.gov/rest/pug"
    text_types = str, bytes
    if not identifier:
        raise ValueError("identifier/cid cannot be None")
    # If identifier is a list, join with commas into string
    if isinstance(identifier, int):
        identifier = str(identifier)
    if not isinstance(identifier, text_types):
        identifier = ",".join(str(x) for x in identifier)

    # Build API URL
    urlid, postdata = None, None
    if namespace == "sourceid":
        identifier = identifier.replace("/", ".")
    if (namespace in ["listkey", "formula", "sourceid"] or
            searchtype == "xref" or (searchtype and namespace == "cid") or
            domain == "sources"):
        urlid = quote(identifier.encode("utf8"))
    else:
        postdata = urlencode([(namespace, identifier)]).encode("utf8")
    comps = filter(
        None,
        [API_BASE, domain, searchtype, namespace, urlid, operation, output])
    apiurl = "/".join(comps)
    # Make request
    response = urlopen(apiurl, postdata)
    return response


def uniprot2seq(ProteinID):
    """Get protein sequence from Uniprot ID

    Args:
        ProteinID (str): the uniprot ID

    Returns:
        str: amino acid sequence
    """
    import urllib
    import string
    import urllib.request as ur

    ID = str(ProteinID)
    localfile = ur.urlopen("http://www.uniprot.org/uniprot/" + ID + ".fasta")
    temp = localfile.readlines()
    res = ""
    for i in range(1, len(temp)):
        res = res + temp[i].strip().decode("utf-8")
    return res


def cid2smiles(cid):
    """SMILES string from PubChem CID

    Args:
        cid (str): PubChem CID

    Returns:
        str: SMILES string
    """
    try:
        smiles = _parse_prop(
            {
                "label": "SMILES",
                "name": "Canonical"
            },
            json.loads(
                request(cid).read().decode())["PC_Compounds"][0]["props"],
        )
    except:
        print("cid " + str(cid) + " failed, use NULL string")
        smiles = "NULL"
    return smiles



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from ..utils.load import download_wrapper, pd_load
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class TCREpitopeBinding(multi_pred_dataset.DataLoader):
    """Data loader class to load datasets in T cell receptor (TCR) Specificity Prediction Task.
    More info:

    Task Description: Given the TCR and epitope sequence, predict binding probability.

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False

    """

    def __init__(self, name, path="./data", print_stats=False):
        """Create TCR Specificity Prediction dataloader object"""
        super().__init__(name,
                         path,
                         print_stats,
                         dataset_names=dataset_names["TCREpitopeBinding"])
        self.entity1_name = "TCR"
        self.entity2_name = "Epitope"

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import (
    dataset2target_lists,
    fuzzy_search,
    interaction_dataset_load,
    label_transform,
    NegSample,
    install,
    create_fold,
    create_fold_setting_cold,
    create_combination_split,
    create_fold_time,
    print_sys,
)


class DataLoader(base_dataset.DataLoader):
    """A base data loader class that each bi-instance prediction task dataloader class can inherit from.

    Attributes: TODO

    """

    def __init__(self,
                 name,
                 path,
                 label_name,
                 print_stats,
                 dataset_names,
                 data_config=None):
        """Create a base dataloader object that each multi-instance prediction task dataloader class can inherit from.

        Args:
            name (str): name of dataloader
            path (str): the path where data is saved
            label_name (str): name of label
            print_stats (bool): whether to print statistics of dataset
            dataset_names (str): A list of dataset names available for a task

        Raises:
            ValueError: label name is not available
        """
        if name.lower() in dataset2target_lists.keys():
            # print_sys("Tip: Use tdc.utils.retrieve_label_name_list(
            # '" + name.lower() + "') to retrieve all available label names.")
            if label_name is None:
                raise ValueError(
                    "Please select a label name. "
                    "You can use tdc.utils.retrieve_label_name_list('" +
                    name.lower() + "') to retrieve all available label names.")

        name = fuzzy_search(name, dataset_names)
        if name == "bindingdb_patent":
            aux_column = "Year"
        else:
            aux_column = None

        (entity1, entity2, raw_y, entity1_idx, entity2_idx, aux_column_val, df,
         augment_df) = interaction_dataset_load(name,
                                                path,
                                                label_name,
                                                dataset_names,
                                                aux_column=aux_column,
                                                data_config=data_config)

        self.name = name
        self.entity1 = entity1
        self.entity2 = entity2
        self.raw_y = raw_y
        self.y = raw_y
        self.entity1_idx = entity1_idx
        self.entity2_idx = entity2_idx
        self.path = path
        self.file_format = "csv"
        self.label_name = label_name

        self.entity1_name = "Entity1"
        self.entity2_name = "Entity2"
        self.aux_column = aux_column
        self.aux_column_val = aux_column_val
        self.log_flag = False
        self.two_types = False
        self.augment_df = augment_df
        self.df = df if augment_df else None

    def get_data(self, format="df"):
        """generate data in some format, e.g., pandas.DataFrame

        Args:
            format (str, optional):
                format of data, the default value is 'df' (DataFrame)

        Returns:
            pandas DataFrame/dict: a dataframe of a dataset/a dictionary for key information in the dataset

        Raises:
            AttributeError: Use the correct format input (df, dict, DeepPurpose)
        """
        if format == "df":
            out = None
            if self.aux_column is None:
                out = pd.DataFrame({
                    self.entity1_name + "_ID": self.entity1_idx,
                    self.entity1_name: self.entity1,
                    self.entity2_name + "_ID": self.entity2_idx,
                    self.entity2_name: self.entity2,
                    "Y": self.y,
                })
            else:
                out = pd.DataFrame({
                    self.entity1_name + "_ID": self.entity1_idx,
                    self.entity1_name: self.entity1,
                    self.entity2_name + "_ID": self.entity2_idx,
                    self.entity2_name: self.entity2,
                    "Y": self.y,
                    self.aux_column: self.aux_column_val,
                })
            if not self.augment_df:
                return out
            for col in self.df.columns:
                out[col] = self.df[col].values
            return out

        elif format == "DeepPurpose":
            return self.entity1.values, self.entity2.values, self.y.values
        elif format == "dict":
            out = {
                self.entity1_name + "_ID": self.entity1_idx.values,
                self.entity1_name: self.entity1.values,
                self.entity2_name + "_ID": self.entity2_idx.values,
                self.entity2_name: self.entity2.values,
                "Y": self.y.values,
            }
            if not self.augment_df:
                return out
            for col in self.df.columns:
                out[col] = self.df[col].values
            return out
        else:
            raise AttributeError("Please use the correct format input")

    def print_stats(self):
        """print the statistics of the dataset"""
        print_sys("--- Dataset Statistics ---")
        try:
            x = np.unique(self.entity1)
        except:
            x = np.unique(self.entity1_idx)

        try:
            y = np.unique(self.entity2)
        except:
            y = np.unique(self.entity2_idx)

        print(
            str(len(x)) + " unique " + self.entity1_name.lower() + "s.",
            flush=True,
            file=sys.stderr,
        )
        print(
            str(len(y)) + " unique " + self.entity2_name.lower() + "s.",
            flush=True,
            file=sys.stderr,
        )
        print(
            str(len(self.y)) + " " + self.entity1_name.lower() + "-" +
            self.entity2_name.lower() + " pairs.",
            flush=True,
            file=sys.stderr,
        )
        print_sys("--------------------------")

    def get_split(
        self,
        method="random",
        seed=42,
        frac=[0.7, 0.1, 0.2],
        column_name=None,
        time_column=None,
    ):
        """split dataset into train/validation/test.

        Args:
            method (str, optional):
                split method, the default value is 'random'
            seed (int, optional):
                random seed, defaults to '42'
            frac (list, optional):
                train/val/test split fractions, defaults to '[0.7, 0.1, 0.2]'
            column_name (Optional[Union[str, List[str]]]): Optional column name(s) to
                split on for cold splits. Defaults to None.
            time_column (None, optional): Description

        Returns:
            dict: a dictionary with three keys ('train', 'valid', 'test'), each value
            is a pandas dataframe object of the splitted dataset.

        Raises:
            AttributeError: the input split method is not available.

        """
        df = self.get_data(format="df")

        if isinstance(column_name, str):
            column_name = [column_name]

        if method == "random":
            return create_fold(df, seed, frac)
        elif method == "cold_" + self.entity1_name.lower():
            return create_fold_setting_cold(df, seed, frac, self.entity1_name)
        elif method == "cold_" + self.entity2_name.lower():
            return create_fold_setting_cold(df, seed, frac, self.entity2_name)
        elif method == "cold_split":
            if column_name is None or not all(
                    list(map(lambda x: x in df.columns.values, column_name))):
                raise AttributeError(
                    "For cold_split, please provide one or multiple column names "
                    "that are contained in the dataframe.")
            return create_fold_setting_cold(df, seed, frac, column_name)
        elif method == "combination":
            return create_combination_split(df, seed, frac)
        elif method == "time":
            if time_column is None:
                raise ValueError(
                    "Please specify the column that has the time variable using time_column."
                )
            return create_fold_time(df, frac, time_column)

        else:
            raise AttributeError(
                "Please select method from random, time, combination or cold_split."
            )

    def neg_sample(self, frac=1):
        """negative sampling

        Args:
            frac (int, optional): the ratio between negative and positive samples.

        Returns:
            DataLoader, the class itself.
        """
        df = NegSample(
            df=self.get_data(format="df"),
            column_names=[
                self.entity1_name + "_ID",
                self.entity1_name,
                self.entity2_name + "_ID",
                self.entity2_name,
            ],
            frac=frac,
            two_types=self.two_types,
        )
        self.entity1_idx = df[self.entity1_name + "_ID"]
        self.entity2_idx = df[self.entity2_name + "_ID"]
        self.entity1 = df[self.entity1_name]
        self.entity2 = df[self.entity2_name]
        self.y = df["Y"]
        self.raw_y = self.y
        return self

    def to_graph(
        self,
        threshold=None,
        format="edge_list",
        split=True,
        frac=[0.7, 0.1, 0.2],
        seed=42,
        order="descending",
    ):
        """Summary TODO

        Args:
            threshold (float, optional): threshold to binarize the data.
            format (str, optional): format of data, defaults to 'edge_list'
            split (bool, optional): if we need to split data into train/valid/test.
            frac (list, optional):  train/val/test split fractions, defaults to '[0.7, 0.1, 0.2]'
            seed (int, optional):  random seed, defaults to '42'
            order (str, optional): order of label transform

        Returns:
            dict: a dictionary for key information in the dataset

        Raises:
            AttributeError: the threshold is not available.
            ImportError: install the required package
        """
        df = self.get_data(format="df")

        if len(np.unique(self.raw_y)) > 2:
            print(
                "The dataset label consists of affinity scores. "
                "Binarization using threshold " + str(threshold) +
                " is conducted to construct the positive edges in the network. "
                "Adjust the threshold by to_graph(threshold = X)",
                flush=True,
                file=sys.stderr,
            )
            if threshold is None:
                raise AttributeError(
                    "Please specify the threshold to binarize the data by "
                    "'to_graph(threshold = N)'!")
            df["label_binary"] = label_transform(self.raw_y,
                                                 True,
                                                 threshold,
                                                 False,
                                                 verbose=False,
                                                 order=order)
        else:
            # already binary
            df["label_binary"] = df["Y"]

        df[self.entity1_name + "_ID"] = df[self.entity1_name +
                                           "_ID"].astype(str)
        df[self.entity2_name + "_ID"] = df[self.entity2_name +
                                           "_ID"].astype(str)
        df_pos = df[df.label_binary == 1]
        df_neg = df[df.label_binary == 0]

        return_dict = {}

        pos_edges = df_pos[[
            self.entity1_name + "_ID", self.entity2_name + "_ID"
        ]].values
        neg_edges = df_neg[[
            self.entity1_name + "_ID", self.entity2_name + "_ID"
        ]].values
        edges = df[[self.entity1_name + "_ID",
                    self.entity2_name + "_ID"]].values

        if format == "edge_list":
            return_dict["edge_list"] = pos_edges
            return_dict["neg_edges"] = neg_edges
        elif format == "dgl":
            try:
                import dgl
            except:
                install("dgl")
                import dgl
            unique_entities = np.unique(pos_edges.T.flatten()).tolist()
            index = list(range(len(unique_entities)))
            dict_ = dict(zip(unique_entities, index))
            edge_list1 = np.array([dict_[i] for i in pos_edges.T[0]])
            edge_list2 = np.array([dict_[i] for i in pos_edges.T[1]])
            return_dict["dgl_graph"] = dgl.DGLGraph((edge_list1, edge_list2))
            return_dict["index_to_entities"] = dict_

        elif format == "pyg":
            try:
                import torch
                from torch_geometric.data import Data
            except:
                raise ImportError(
                    "Please see https://pytorch-geometric.readthedocs.io/en/latest/notes/installation.html to install pytorch geometric!"
                )

            unique_entities = np.unique(pos_edges.T.flatten()).tolist()
            index = list(range(len(unique_entities)))
            dict_ = dict(zip(unique_entities, index))
            edge_list1 = np.array([dict_[i] for i in pos_edges.T[0]])
            edge_list2 = np.array([dict_[i] for i in pos_edges.T[1]])

            edge_index = torch.tensor([edge_list1, edge_list2],
                                      dtype=torch.long)
            x = torch.tensor(np.array(index), dtype=torch.float)
            data = Data(x=x, edge_index=edge_index)
            return_dict["pyg_graph"] = data
            return_dict["index_to_entities"] = dict_

        elif format == "df":
            return_dict["df"] = df

        if split:
            return_dict["split"] = create_fold(df, seed, frac)

        return return_dict



import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .utils import fuzzy_search
from .metadata import evaluator_name, distribution_oracles

try:
    from sklearn.metrics import (
        roc_auc_score,
        f1_score,
        average_precision_score,
        precision_score,
        recall_score,
        accuracy_score,
        precision_recall_curve,
    )
    from sklearn.metrics import (
        mean_squared_error,
        mean_absolute_error,
        r2_score,
        cohen_kappa_score,
        auc,
        roc_curve,
    )
except:
    ImportError(
        "Please install sklearn by 'conda install -c anaconda scikit-learn' or 'pip install scikit-learn '! "
    )


def avg_auc(y_true, y_pred):
    scores = []
    for i in range(np.array(y_true).shape[0]):
        scores.append(roc_auc_score(y_true[i], y_pred[i]))
    return sum(scores) / len(scores)


def rmse(y_true, y_pred):
    return np.sqrt(mean_squared_error(y_true, y_pred))


def recall_at_precision_k(y_true, y_pred, threshold=0.9):
    pr, rc, thr = precision_recall_curve(y_true, y_pred)
    if len(np.where(pr >= threshold)[0]) > 0:
        return rc[np.where(pr >= threshold)[0][0]]
    else:
        return 0.0


def precision_at_recall_k(y_true, y_pred, threshold=0.9):
    pr, rc, thr = precision_recall_curve(y_true, y_pred)
    if len(np.where(rc >= threshold)[0]) > 0:
        return pr[np.where(rc >= threshold)[0][-1]]
    else:
        return 0.0


def pcc(y_true, y_pred):
    return np.corrcoef(y_true, y_pred)[1, 0]


def range_logAUC(true_y, predicted_score, FPR_range=(0.001, 0.1)):
    """
    Author: Yunchao "Lance" Liu (lanceknight26@gmail.com)
    Calculate logAUC in a certain FPR range (default range: [0.001, 0.1]).
    This was used by previous methods [1] and the reason is that only a
    small percentage of samples can be selected for experimental tests in
    consideration of cost. This means only molecules with very high
    predicted score can be worth testing, i.e., the decision
    threshold is high. And the high decision threshold corresponds to the
    left side of the ROC curve, i.e., those FPRs with small values. Also,
    because the threshold cannot be predetermined, the area under the curve
    is used to consolidate all possible thresholds within a certain FPR
    range. Finally, the logarithm is used to bias smaller FPRs. The higher
    the logAUC[0.001, 0.1], the better the performance.

    A perfect classifer gets a logAUC[0.001, 0.1] ) of 1, while a random
    classifer gets a logAUC[0.001, 0.1] ) of around 0.0215 (See [2])

    References:
    [1] Mysinger, M.M. and B.K. Shoichet, Rapid Context-Dependent Ligand
    Desolvation in Molecular Docking. Journal of Chemical Information and
    Modeling, 2010. 50(9): p. 1561-1573.
    [2] Liu, Yunchao, et al. "Interpretable Chirality-Aware Graph Neural
    Network for Quantitative Structure Activity Relationship Modeling in
    Drug Discovery." bioRxiv (2022).
    :param true_y: numpy array of the ground truth. Values are either 0
    (inactive) or 1(active).
    :param predicted_score: numpy array of the predicted score (The
    score does not have to be between 0 and 1)
    :param FPR_range: the range for calculating the logAUC formated in
    (x, y) with x being the lower bound and y being the upper bound
    :return: a numpy array of logAUC of size [1,1]
    """

    # FPR range validity check
    if FPR_range == None:
        raise Exception("FPR range cannot be None")
    lower_bound = FPR_range[0]
    upper_bound = FPR_range[1]
    if lower_bound >= upper_bound:
        raise Exception("FPR upper_bound must be greater than lower_bound")

    fpr, tpr, thresholds = roc_curve(true_y, predicted_score, pos_label=1)

    tpr = np.append(tpr, np.interp([lower_bound, upper_bound], fpr, tpr))
    fpr = np.append(fpr, [lower_bound, upper_bound])

    # Sort both x-, y-coordinates array
    tpr = np.sort(tpr)
    fpr = np.sort(fpr)

    # Get the data points' coordinates. log_fpr is the x coordinate, tpr is the y coordinate.
    log_fpr = np.log10(fpr)
    x = log_fpr
    y = tpr
    lower_bound = np.log10(lower_bound)
    upper_bound = np.log10(upper_bound)

    # Get the index of the lower and upper bounds
    lower_bound_idx = np.where(x == lower_bound)[-1][-1]
    upper_bound_idx = np.where(x == upper_bound)[-1][-1]

    # Create a new array trimmed at the lower and upper bound
    trim_x = x[lower_bound_idx:upper_bound_idx + 1]
    trim_y = y[lower_bound_idx:upper_bound_idx + 1]

    area = auc(trim_x, trim_y) / (upper_bound - lower_bound)
    return area


## code source check here https://github.com/charnley/rmsd


def centroid(X):
    """
    Centroid is the mean position of all the points in all of the coordinate
    directions, from a vectorset X.
    https://en.wikipedia.org/wiki/Centroid
    C = sum(X)/len(X)
    Parameters
    ----------
    X : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    C : float
        centroid
    """
    C = X.mean(axis=0)
    return C


def rmsd(V, W):
    """
    Calculate Root-mean-square deviation from two sets of vectors V and W.
    Parameters
    ----------
    V : array
        (N,D) matrix, where N is points and D is dimension.
    W : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    rmsd : float
        Root-mean-square deviation between the two vectors
    """
    diff = np.array(V) - np.array(W)
    N = len(V)
    return np.sqrt((diff * diff).sum() / N)


def kabsch(P, Q):
    """
    Using the Kabsch algorithm with two sets of paired point P and Q, centered
    around the centroid. Each vector set is represented as an NxD
    matrix, where D is the the dimension of the space.
    The algorithm works in three steps:
    - a centroid translation of P and Q (assumed done before this function
      call)
    - the computation of a covariance matrix C
    - computation of the optimal rotation matrix U
    For more info see http://en.wikipedia.org/wiki/Kabsch_algorithm
    Parameters
    ----------
    P : array
        (N,D) matrix, where N is points and D is dimension.
    Q : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    U : matrix
        Rotation matrix (D,D)
    """

    # Computation of the covariance matrix
    C = np.dot(np.transpose(P), Q)

    # Computation of the optimal rotation matrix
    # This can be done using singular value decomposition (SVD)
    # Getting the sign of the det(V)*(W) to decide
    # whether we need to correct our rotation matrix to ensure a
    # right-handed coordinate system.
    # And finally calculating the optimal rotation matrix U
    # see http://en.wikipedia.org/wiki/Kabsch_algorithm
    V, S, W = np.linalg.svd(C)
    d = (np.linalg.det(V) * np.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    # Create Rotation matrix U
    U = np.dot(V, W)

    return U


def kabsch_rmsd(P, Q, W=None, translate=False):
    """
    Rotate matrix P unto Q using Kabsch algorithm and calculate the RMSD.
    An optional vector of weights W may be provided.
    Parameters
    ----------
    P : array
        (N,D) matrix, where N is points and D is dimension.
    Q : array
        (N,D) matrix, where N is points and D is dimension.
    W : array or None
        (N) vector, where N is points.
    translate : bool
        Use centroids to translate vector P and Q unto each other.
    Returns
    -------
    rmsd : float
        root-mean squared deviation
    """

    if translate:
        Q = Q - centroid(Q)
        P = P - centroid(P)

    if W is not None:
        return kabsch_weighted_rmsd(P, Q, W)

    P = kabsch_rotate(P, Q)
    return rmsd(P, Q)


def kabsch_rotate(P, Q):
    """
    Rotate matrix P unto matrix Q using Kabsch algorithm.
    Parameters
    ----------
    P : array
        (N,D) matrix, where N is points and D is dimension.
    Q : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    P : array
        (N,D) matrix, where N is points and D is dimension,
        rotated
    """
    U = kabsch(P, Q)

    # Rotate P
    P = np.dot(P, U)
    return P


def kabsch_weighted_rmsd(P, Q, W=None):
    """
    Calculate the RMSD between P and Q with optional weighhts W
    Parameters
    ----------
    P : array
        (N,D) matrix, where N is points and D is dimension.
    Q : array
        (N,D) matrix, where N is points and D is dimension.
    W : vector
        (N) vector, where N is points
    Returns
    -------
    RMSD : float
    """
    R, T, w_rmsd = kabsch_weighted(P, Q, W)
    return w_rmsd


def kabsch_weighted(P, Q, W=None):
    """
    Using the Kabsch algorithm with two sets of paired point P and Q.
    Each vector set is represented as an NxD matrix, where D is the
    dimension of the space.
    An optional vector of weights W may be provided.
    Note that this algorithm does not require that P and Q have already
    been overlayed by a centroid translation.
    The function returns the rotation matrix U, translation vector V,
    and RMS deviation between Q and P', where P' is:
        P' = P * U + V
    For more info see http://en.wikipedia.org/wiki/Kabsch_algorithm
    Parameters
    ----------
    P : array
        (N,D) matrix, where N is points and D is dimension.
    Q : array
        (N,D) matrix, where N is points and D is dimension.
    W : array or None
        (N) vector, where N is points.
    Returns
    -------
    U    : matrix
           Rotation matrix (D,D)
    V    : vector
           Translation vector (D)
    RMSD : float
           Root mean squared deviation between P and Q
    """
    # Computation of the weighted covariance matrix
    CMP = np.zeros(3)
    CMQ = np.zeros(3)
    C = np.zeros((3, 3))
    if W is None:
        W = np.ones(len(P)) / len(P)
    W = np.array([W, W, W]).T
    # NOTE UNUSED psq = 0.0
    # NOTE UNUSED qsq = 0.0
    iw = 3.0 / W.sum()
    n = len(P)
    for i in range(3):
        for j in range(n):
            for k in range(3):
                C[i, k] += P[j, i] * Q[j, k] * W[j, i]
    CMP = (P * W).sum(axis=0)
    CMQ = (Q * W).sum(axis=0)
    PSQ = (P * P * W).sum() - (CMP * CMP).sum() * iw
    QSQ = (Q * Q * W).sum() - (CMQ * CMQ).sum() * iw
    C = (C - np.outer(CMP, CMQ) * iw) * iw

    # Computation of the optimal rotation matrix
    # This can be done using singular value decomposition (SVD)
    # Getting the sign of the det(V)*(W) to decide
    # whether we need to correct our rotation matrix to ensure a
    # right-handed coordinate system.
    # And finally calculating the optimal rotation matrix U
    # see http://en.wikipedia.org/wiki/Kabsch_algorithm
    V, S, W = np.linalg.svd(C)
    d = (np.linalg.det(V) * np.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    # Create Rotation matrix U, translation vector V, and calculate RMSD:
    U = np.dot(V, W)
    msd = (PSQ + QSQ) * iw - 2.0 * S.sum()
    if msd < 0.0:
        msd = 0.0
    rmsd_ = np.sqrt(msd)
    V = np.zeros(3)
    for i in range(3):
        t = (U[i, :] * CMQ).sum()
        V[i] = CMP[i] - t
    V = V * iw
    return U, V, rmsd_


class Evaluator:
    """evaluator to evaluate predictions

    Args:
            name (str): the name of the evaluator function
    """

    def __init__(self, name):
        """create an evaluate object"""
        self.name = fuzzy_search(name, evaluator_name)
        self.assign_evaluator()

    def assign_evaluator(self):
        """obtain evaluator function given the evaluator name"""
        if self.name == "roc-auc":
            self.evaluator_func = roc_auc_score
        elif self.name == "f1":
            self.evaluator_func = f1_score
        elif self.name == "pr-auc":
            self.evaluator_func = average_precision_score
        elif self.name == "rp@k":
            self.evaluator_func = recall_at_precision_k
        elif self.name == "pr@k":
            self.evaluator_func = precision_at_recall_k
        elif self.name == "precision":
            self.evaluator_func = precision_score
        elif self.name == "recall":
            self.evaluator_func = recall_score
        elif self.name == "accuracy":
            self.evaluator_func = accuracy_score
        elif self.name == "mse":
            self.evaluator_func = mean_squared_error
        elif self.name == "rmse":
            self.evaluator_func = rmse
        elif self.name == "mae":
            self.evaluator_func = mean_absolute_error
        elif self.name == "r2":
            self.evaluator_func = r2_score
        elif self.name == "pcc":
            self.evaluator_func = pcc
        elif self.name == "spearman":
            try:
                from scipy import stats
            except:
                ImportError("Please install scipy by 'pip install scipy'! ")
            self.evaluator_func = stats.spearmanr
        elif self.name == "micro-f1":
            self.evaluator_func = f1_score
        elif self.name == "macro-f1":
            self.evaluator_func = f1_score
        elif self.name == "kappa":
            self.evaluator_func = cohen_kappa_score
        elif self.name == "avg-roc-auc":
            self.evaluator_func = avg_auc
        elif self.name == "novelty":
            from .chem_utils import novelty

            self.evaluator_func = novelty
        elif self.name == "diversity":
            from .chem_utils import diversity

            self.evaluator_func = diversity
        elif self.name == "validity":
            from .chem_utils import validity

            self.evaluator_func = validity
        elif self.name == "uniqueness":
            from .chem_utils import uniqueness

            self.evaluator_func = uniqueness
        elif self.name == "kl_divergence":
            from .chem_utils import kl_divergence

            self.evaluator_func = kl_divergence
        elif self.name == "fcd_distance":
            from .chem_utils import fcd_distance

            self.evaluator_func = fcd_distance
        elif self.name == "range_logAUC":
            self.evaluator_func = range_logAUC
        elif self.name == "rmsd":
            self.evaluator_func = rmsd
        elif self.name == "kabsch_rmsd":
            self.evaluator_func = kabsch_rmsd

    def __call__(self, *args, **kwargs):
        """call the evaluator function on targets and predictions

        Args:
            *args: targets, predictions, and other information
            **kwargs: other auxilliary inputs for some evaluators

        Returns:
            float: the evaluator output
        """
        if self.name in distribution_oracles:
            return self.evaluator_func(*args, **kwargs)
        # 	#### evaluator for distribution learning, e.g., diversity, validity
        y_true = kwargs["y_true"] if "y_true" in kwargs else args[0]
        y_pred = kwargs["y_pred"] if "y_pred" in kwargs else args[1]
        if len(args) <= 2 and "threshold" not in kwargs:
            threshold = 0.5
        else:
            threshold = kwargs["threshold"] if "threshold" in kwargs else args[2]

        ### original __call__(self, y_true, y_pred, threshold = 0.5)
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        if self.name in ["precision", "recall", "f1", "accuracy"]:
            y_pred = [1 if i > threshold else 0 for i in y_pred]
        if self.name in ["micro-f1", "macro-f1"]:
            return self.evaluator_func(y_true, y_pred, average=self.name[:5])
        if self.name in ["rp@k", "pr@k"]:
            return self.evaluator_func(y_true, y_pred, threshold=threshold)
        if self.name == "spearman":
            return self.evaluator_func(y_true, y_pred)[0]
        return self.evaluator_func(y_true, y_pred)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import sys
import warnings

warnings.filterwarnings("ignore")

from . import single_pred_dataset
from ..utils import print_sys
from ..metadata import dataset_names


class HTS(single_pred_dataset.DataLoader):
    """Data loader class to load datasets in HTS task. More info: https://tdcommons.ai/single_pred_tasks/hts/

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False
        convert_format (str, optional):
            Automatic conversion of SMILES to other molecular formats in MolConvert class. Stored as separate column in dataframe, defaults to None
    """

    def __init__(
        self,
        name,
        path="./data",
        label_name=None,
        print_stats=False,
        convert_format=None,
    ):
        """Create a HTS dataloader object."""
        super().__init__(
            name,
            path,
            label_name,
            print_stats,
            dataset_names=dataset_names["HTS"],
            convert_format=convert_format,
        )
        if print_stats:
            self.print_stats()
        print("Done!", flush=True, file=sys.stderr)



from .config import ResourceConfig
from ..feature_generators.resource import ResourceFeatureGenerator


class OpentargetsDTI(ResourceConfig):
    """Configuration for opentargets drug-target-identification datasets"""

    def __init__(self):
        super(OpentargetsDTI, self).__init__(
            ResourceFeatureGenerator(),
            keys=[
                "opentargets_ra_data_splits", "opentargets_ibd_data_splits",
                "df"
            ],
            loader_functions=["split", "split", "concat"],
            loader_args=[{
                "dataset": ("self", "opentargets_ra_drug_evidence"),
                "column_name":
                    "targetId_genename",
                "pos_train": ("self", [
                    "opentargets_ra_data_splits", "splits", "pos_train_names"
                ]),
                "pos_dev":
                    None,
                "pos_test":
                    ("self",
                     ["opentargets_ra_data_splits", "splits",
                      "pos_test_names"]),
                "neg_train": ("self", [
                    "opentargets_ra_data_splits", "splits", "neg_train_names"
                ]),
                "neg_dev":
                    None,
                "neg_test":
                    ("self",
                     ["opentargets_ra_data_splits", "splits",
                      "neg_test_names"]),
            }, {
                "dataset": ("self", "opentargets_ibd_drug_evidence"),
                "column_name":
                    "targetId_genename",
                "pos_train": ("self", [
                    "opentargets_ibd_data_splits", "splits", "pos_train_names"
                ]),
                "pos_dev":
                    None,
                "pos_test": ("self", [
                    "opentargets_ibd_data_splits", "splits", "pos_test_names"
                ]),
                "neg_train": ("self", [
                    "opentargets_ibd_data_splits", "splits", "neg_train_names"
                ]),
                "neg_dev":
                    None,
                "neg_test": ("self", [
                    "opentargets_ibd_data_splits", "splits", "neg_test_names"
                ])
            }, {
                "ds_list": [
                    "opentargets_ibd_data_splits", "opentargets_ra_data_splits"
                ],
                "axis": 0
            }])



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT
import os

from .base_group import BenchmarkGroup
from ..resource.pinnacle import PINNACLE


class SCDTIGroup(BenchmarkGroup):
    """Create SCDTI Group Class object. This is for single-cell drug-target identification task benchmark.

    Args:
            path (str, optional): the path to store/retrieve the SCDTI group datasets.
    """

    def __init__(self, path="./data", file_format="csv"):
        """Create an SCDTI benchmark group class."""
        self.name = "SCDTI_Group"
        self.path = os.path.join(path, self.name)
        self.p = PINNACLE()

    def precision_recall_at_k(self, y, preds, k: int = 5):
        """
        Calculate recall@k and precision@k for binary classification.
        """
        import numpy as np
        import pandas as pd
        from sklearn.metrics import accuracy_score, average_precision_score
        assert preds.shape[0] == y.shape[0]
        assert k > 0
        if k > preds.shape[0]:
            return -1, -1, -1, -1

        # Sort the scores and the labels by the scores
        sorted_indices = np.argsort(preds.flatten())[::-1]
        sorted_preds = preds[sorted_indices]
        sorted_y = y[sorted_indices]

        # Get the scores of the k highest predictions
        topk_preds = sorted_preds[:k]
        topk_y = sorted_y[:k]

        # Calculate the recall@k and precision@k
        recall_k = np.sum(topk_y) / np.sum(y)
        precision_k = np.sum(topk_y) / k

        # Calculate the accuracy@k
        accuracy_k = accuracy_score(topk_y, topk_preds > 0.5)

        # Calculate the AP@k
        ap_k = average_precision_score(topk_y, topk_preds)

        return recall_k, precision_k, accuracy_k, ap_k

    def get_train_valid_split(self, seed=1):
        """parameters included for compatibility. this benchmark has a fixed train/test split."""
        train = self.p.get_exp_data(seed=seed, split="train")
        val = self.p.get_exp_data(seed=seed, split="val")
        return {"train": train, "val": val}

    def get_test(self, seed=1):
        return {"test": self.p.get_exp_data(seed=seed, split="test")}

    def evaluate(self, y_pred, k=5, top_k=20, seed=1):
        from numpy import mean
        from sklearn.metrics import roc_auc_score
        y_true = self.get_test(seed=seed)["test"]
        assert "preds" in y_pred.columns, "require 'preds' prediction label in input df"
        assert "cell_type_label" in y_pred.columns, "require cell_type_label in input df"
        assert "disease" in y_pred.columns, "require 'disease' in input df"
        cells = y_true["cell_type_label"].unique()
        diseases = y_true["disease"].unique()
        assert len(cells) == len(
            y_pred["cell_type_label"].unique()
        ), "number of cell types in input df and test df do not match. expected {}".format(
            len(cells))
        assert len(diseases) == len(
            y_pred["disease"].unique()
        ), "number of diseases in input df do not match test df. expected {}".format(
            len(diseases))
        results = {d: [] for d in diseases}
        for disease in diseases:
            for cell in cells:
                preds = y_pred[(y_pred["disease"] == disease) &
                               (y_pred["cell_type_label"] == cell)]
                yt = y_true[(y_true["disease"] == disease) &
                            (y_true["cell_type_label"] == cell)]
                assert len(preds) == len(
                    yt
                ), "mismatch in length of predictions and results for a specific disease {} and cell type {}".format(
                    disease, cell)
                if len(yt) == 0:
                    continue
                auc = roc_auc_score(yt["y"], preds["preds"])
                recall_k, precision_k, accuracy_k, ap_k = self.precision_recall_at_k(
                    yt["y"].values, preds["preds"].values, k=k)
                results[disease].append({
                    "auc": auc,
                    "recall": recall_k,
                    "precision": precision_k,
                    "accuracy": accuracy_k,
                    "ap": ap_k
                })
        # for now, we benchmark with only ap@k with top 20 cells
        for d, scores in results.items():
            assert type(
                scores
            ) == list, "scores should be a list. got {} with value {}".format(
                scores, type(scores))
            assert type(scores[0]
                       ) == dict, "scores should contain dictionary of metrics"
            assert "ap" in scores[0], "scores should include 'ap'"
            topk_cells = [
                x["ap"] for x in sorted(scores, key=lambda s: s["ap"])[-top_k:]
            ]
            results[d] = mean(topk_cells)
        return results

    def evaluate_many(self, preds: list, seed=1):
        from numpy import mean, std
        assert type(
            preds
        ) == list, "expected preds to be a list containing prediction dataframes for multiple seeds"
        if len(preds) < 5:
            raise Exception(
                "Run your model on at least 5 seeds to compare results and provide your outputs in preds."
            )
        evals = [self.evaluate(x, seed=seed) for x in preds]
        diseases = preds[0]["disease"].unique()
        return {
            d: [mean([x[d] for x in evals]),
                std([x[d] for x in evals])] for d in diseases
        }



import sys
from vina import Vina
from time import time

ligand_pdbqt_file = sys.argv[1]
receptor_pdbqt_file = sys.argv[2]
output_file = sys.argv[3]
center = [sys.argv[4], sys.argv[5], sys.argv[6]]
center = [float(i) for i in center]
box_size = [sys.argv[7], sys.argv[8], sys.argv[9]]
box_size = [float(i) for i in box_size]


# print(ligand_pdbqt_file, receptor_pdbqt_file, output_file, center, box_size)
def docking(ligand_pdbqt_file, receptor_pdbqt_file, output_file, center,
            box_size):
    t1 = time()
    v = Vina(sf_name="vina")
    v.set_receptor(rigid_pdbqt_filename=receptor_pdbqt_file)
    v.set_ligand_from_file(ligand_pdbqt_file)
    v.compute_vina_maps(center=center, box_size=box_size)
    energy = v.score()
    energy_minimized = v.optimize()
    t2 = time()
    print("vina takes seconds: ", str(t2 - t1)[:5])
    with open(output_file, "w") as fout:
        fout.write(str(energy_minimized[0]))


docking(ligand_pdbqt_file, receptor_pdbqt_file, output_file, center, box_size)
"""
Example: 
    python XXXX.py data/1iep_ligand.pdbqt ./data/1iep_receptor.pdbqt ./data/out 15.190 53.903 16.917 20 20 20 
"""



from functools import wraps
import inspect
import pandas as pd

from .data_feature_generator import DataFeatureGenerator


def parse_args(func):

    def check(self, *args, **kwargs):
        for idx, arg in enumerate(args):
            if isinstance(arg, tuple) and arg[0] == "self" and len(arg) == 2:
                if type(arg[1]) not in [list]:
                    args[idx] = args[0][
                        arg[1]]  # convert to variable in the DataLoader
                else:
                    tmp = args[0]
                    for a in arg[1]:
                        tmp = tmp[a]
                    args[idx] = tmp

        for k, v in kwargs.items():
            if isinstance(v, tuple) and v[0] == "self" and len(v) == 2:
                if type(v[1]) not in [list]:
                    try:
                        kwargs[k] = kwargs["dataloader"][
                            v[1]]  # convert to variable in the DataLoader
                    except:
                        raise Exception(
                            "exception assigning kwargs",
                            k,
                            v,
                            kwargs["dataloader"].keys(),
                        )
                else:
                    tmp = kwargs["dataloader"]
                    for a in v[1]:
                        tmp = tmp[a]
                    kwargs[k] = tmp
        return func(self, *args, **kwargs)

    return check


class ResourceFeatureGenerator(DataFeatureGenerator):

    @parse_args
    def join(self,
             dataloader,
             variable_to_update,
             left,
             right,
             on_left,
             on_right=None,
             spec=None):
        spec = spec or "inner"
        assert spec in ["inner", "outer", "left", "right"
                       ], "spec must be one of inner, outer, left, right"
        assert isinstance(left, pd.DataFrame)
        assert isinstance(right, pd.DataFrame)
        assert isinstance(on_left, list)
        if on_right:
            assert isinstance(on_right, list)
            assert len(on_right) == len(on_left)
            for idx, col in enumerate(on_right):
                right[on_left[idx]] = on_right[
                    col]  # rename columns to match the ones specified
        if spec == "inner":
            dataloader[variable_to_update] = pd.merge(left,
                                                      right,
                                                      on=on_left,
                                                      how="inner")
        elif spec == "left":
            raise ValueError("Left joins are not supported.")
        elif spec == "right":
            raise ValueError("Right joins are not supported.")
        elif spec == "outer":
            raise ValueError("Outer joins are not supported.")
        return dataloader[variable_to_update]

    @parse_args
    def variable_plug(self, dataloader, variable_to_update, column_name, value):
        dataloader[variable_to_update][column_name] = value
        return dataloader[variable_to_update]

    @parse_args
    def split(self,
              dataloader=None,
              variable_to_update=None,
              dataset=None,
              column_name=None,
              pos_train=None,
              pos_dev=None,
              pos_test=None,
              neg_train=None,
              neg_dev=None,
              neg_test=None):
        if pos_train is not None:
            # raise Exception("dataset", dataset.columns, len(dataset), max(pos_train.values))
            dataloader[variable_to_update]["train_split_pos"] = dataset[
                dataset[column_name].isin(pos_train)]
            dataloader[variable_to_update]["train_split_pos"]["Y"] = 1
        if pos_dev is not None:
            dataloader[variable_to_update]["dev_split_pos"] = dataset[
                dataset[column_name].isin(pos_dev)]
            dataloader[variable_to_update]["dev_split_pos"]["Y"] = 1
        if pos_test is not None:
            dataloader[variable_to_update]["test_split_pos"] = dataset[
                dataset[column_name].isin(pos_test)]
            dataloader[variable_to_update]["test_split_pos"]["Y"] = 1
        if neg_train is not None:
            dataloader[variable_to_update]["train_split_neg"] = dataset[
                dataset[column_name].isin(neg_train)]
            dataloader[variable_to_update]["train_split_neg"]["Y"] = 0
        if neg_dev is not None:
            dataloader[variable_to_update]["dev_split_neg"] = dataset[
                dataset[column_name].isin(neg_dev)]
            dataloader[variable_to_update]["dev_split_neg"]["Y"] = 0
        if neg_test is not None:
            dataloader[variable_to_update]["test_split_neg"] = dataset[
                dataset[column_name].isin(neg_test)]
            dataloader[variable_to_update]["test_split_neg"]["Y"] = 0
        df = pd.DataFrame(columns=dataset.columns)
        splits_dict = {
            "train": pd.DataFrame(columns=dataset.columns),
            "dev": pd.DataFrame(columns=dataset.columns),
            "test": pd.DataFrame(columns=dataset.columns)
        }
        if pos_train is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["train_split_pos"]], axis=0)
            splits_dict["train"] = pd.concat([
                splits_dict["train"],
                dataloader[variable_to_update]["train_split_pos"]
            ],
                                             axis=0)
        if pos_dev is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["dev_split_pos"]], axis=0)
            splits_dict["dev"] = pd.concat([
                splits_dict["dev"],
                dataloader[variable_to_update]["dev_split_pos"]
            ],
                                           axis=0)
        if pos_test is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["test_split_pos"]], axis=0)
            splits_dict["test"] = pd.concat([
                splits_dict["test"],
                dataloader[variable_to_update]["test_split_pos"]
            ],
                                            axis=0)
        if neg_train is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["train_split_neg"]], axis=0)
            splits_dict["train"] = pd.concat([
                splits_dict["train"],
                dataloader[variable_to_update]["train_split_neg"]
            ],
                                             axis=0)
        if neg_dev is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["dev_split_neg"]], axis=0)
            splits_dict["dev"] = pd.concat([
                splits_dict["dev"],
                dataloader[variable_to_update]["dev_split_neg"]
            ],
                                           axis=0)
        if neg_test is not None:
            df = pd.concat(
                [df, dataloader[variable_to_update]["test_split_neg"]], axis=0)
            splits_dict["test"] = pd.concat([
                splits_dict["test"],
                dataloader[variable_to_update]["test_split_neg"]
            ],
                                            axis=0)
        dataloader[variable_to_update]["data"] = df
        dataloader['split'] = splits_dict
        return dataloader[variable_to_update]

    def process_loader(self, loader, keys, functions, args):
        """Run a series of modifications to a dataloader's variables."""
        functions_dict = {
            name: getattr(self, name) for name, _ in inspect.getmembers(self)
        }
        for idx, f in enumerate(functions):
            a = args[idx]
            a["dataloader"] = loader
            a["variable_to_update"] = keys[idx]
            # a = (loader, keys[idx])
            try:
                functions_dict[f](**a)
            except KeyError as e:
                raise ValueError("Unknown function '{}'".format(f),
                                 functions_dict.keys(), e)
            except Exception as e:
                print("Error in function {} with args {}".format(f, a))
                raise e
        return loader

    @parse_args
    def concat(self, dataloader, variable_to_update, ds_list, axis=None):
        """Concatenate two datasets along some axis."""
        dataloader[variable_to_update] = pd.concat(
            [dataloader[x]["data"] for x in ds_list], axis=axis)
        return dataloader[variable_to_update]



"""Utilities functions for dataset/metadata retrieval
"""
import os, sys
import pandas as pd
from .label_name_list import dataset2target_lists
from .misc import fuzzy_search
from .load import pd_load
from ..metadata import dataset_names, benchmark_names, dataset_list


def get_label_map(
    name,
    path="./data",
    target=None,
    file_format="csv",
    output_format="dict",
    task="DDI",
    name_column="Map",
):
    """to retrieve the biomedical meaning of labels

    Args:
        name (str): the name of the dataset
        path (str, optional): the dataset path, where the data is located
        target (None, optional): the label name
        file_format (str, optional): format of the file
        output_format (str, optional): return a dictionary or a dataframe or the raw array of mapped labels
        task (str, optional): the name of the task
        name_column (str, optional): the name of the column that stores the label name

    Returns:
        dict/pd.DataFrame/np.array: when output_format is dict/df/array

    Raises:
        ValueError: output_format not supported.
    """
    name = fuzzy_search(name, dataset_names[task])
    if target is None:
        target = "Y"
    df = pd_load(name, path)

    if output_format == "dict":
        return dict(zip(df[target].values, df[name_column].values))
    elif output_format == "df":
        return df
    elif output_format == "array":
        return df[name_column].values
    else:
        raise ValueError(
            "Please use the correct output format, select from dict, df, array."
        )


def get_reaction_type(name, path="./data", output_format="array"):
    """to retrieve the type of reactions for reaction dataset

    Args:
        name (str): dataset name
        path (str, optional): dataset path
        output_format (str, optional): output format in dataframe or in raw array format

    Returns:
        pd.DataFrame/np.array: when output_format is df/array

    Raises:
        ValueError: the output format is not supported
    """
    name = fuzzy_search(name, dataset_names["RetroSyn"])
    df = pd_load(name, path)

    if output_format == "df":
        return df
    elif output_format == "array":
        return df["category"].values
    else:
        raise ValueError(
            "Please use the correct output format, select from df, array.")


def retrieve_label_name_list(name):
    """get the set of available labels for query dataset

    Args:
        name (str): rough dataset name

    Returns:
        list: a list of available labels
    """
    name = fuzzy_search(name, dataset_list)
    return dataset2target_lists[name]


def retrieve_dataset_names(name):
    """to get all available dataset names given a task

    Args:
        name (str): the name of query task

    Returns:
        list: a list of available datasets
    """
    return dataset_names[name]


def retrieve_all_benchmarks():
    """to get all available benchmark groups

    Returns:
        list: a list of benchmark group names
    """
    return list(benchmark_names.keys())


def retrieve_benchmark_names(name):
    """to get all available benchmarks given a query benchmark group

    Args:
        name (str): the name of the benchmark group

    Returns:
        list: a list of benchmarks
    """
    name = fuzzy_search(name, list(benchmark_names.keys()))
    datasets = benchmark_names[name]

    dataset_names = []

    for task, datasets in datasets.items():
        for dataset in datasets:
            dataset_names.append(dataset)
    return dataset_names



import numpy as np
import scipy.sparse as sp

from ...utils.load import pd_load, download_wrapper


class GeneformerTokenizer:
    """
    Uses Geneformer Utils to parse zero-shot model server requests for tokenizing single-cell gene expression data.

    Tokenizer source code: https://github.com/amva13/geneformer/blob/main/geneformer/tokenizer.py
    """

    def __init__(
        self,
        path=None,
        custom_attr_name_dict=None,
        nproc=1,
    ):
        path = path or "./data"
        download_wrapper("geneformer_gene_median_dictionary", path,
                         ["geneformer_gene_median_dictionary"])
        download_wrapper("geneformer_gene_name_id_dict", path,
                         ["geneformer_gene_name_id_dict"])
        download_wrapper("geneformer_token_dictionary", path,
                         ["geneformer_token_dictionary"])
        self.gene_median_dict = pd_load("geneformer_gene_median_dictionary",
                                        path=path)
        self.gene_name_id_dict = pd_load("geneformer_gene_name_id_dict",
                                         path=path)
        self.gene_token_dict = pd_load("geneformer_token_dictionary", path=path)
        self.custom_attr_name_dict = custom_attr_name_dict
        self.nproc = nproc

        # gene keys for full vocabulary
        self.gene_keys = list(self.gene_median_dict.keys())

        # protein-coding and miRNA gene list dictionary for selecting .loom rows for tokenization
        self.genelist_dict = dict(
            zip(self.gene_keys, [True] * len(self.gene_keys)))

    @classmethod
    def rank_genes(cls, gene_vector, gene_tokens):
        """
        Rank gene expression vector.
        """
        # sort by median-scaled gene values
        sorted_indices = np.argsort(-gene_vector)
        return gene_tokens[sorted_indices]

    def tokenize_cell_vectors(self,
                              cell_vector_adata,
                              target_sum=10_000,
                              chunk_size=512,
                              ensembl_id="ensembl_id",
                              ncounts="ncounts"):
        """
        Tokenizing single-cell gene expression vectors formatted as anndata types.

        """
        adata = cell_vector_adata
        if self.custom_attr_name_dict is not None:
            file_cell_metadata = {
                attr_key: [] for attr_key in self.custom_attr_name_dict.keys()
            }

        coding_miRNA_loc = np.where([
            self.genelist_dict.get(i, False) for i in adata.var[ensembl_id]
        ])[0]
        norm_factor_vector = np.array([
            self.gene_median_dict[i]
            for i in adata.var[ensembl_id][coding_miRNA_loc]
        ])
        coding_miRNA_ids = adata.var[ensembl_id][coding_miRNA_loc]
        coding_miRNA_tokens = np.array(
            [self.gene_token_dict[i] for i in coding_miRNA_ids])

        try:
            _ = adata.obs["filter_pass"]
        except KeyError:
            var_exists = False
        else:
            var_exists = True

        if var_exists:
            filter_pass_loc = np.where(
                [i == 1 for i in adata.obs["filter_pass"]])[0]
        elif not var_exists:
            print(
                f"The anndata object has no column attribute 'filter_pass'; tokenizing all cells."
            )
            filter_pass_loc = np.array([i for i in range(adata.shape[0])])

        tokenized_cells = []

        for i in range(0, len(filter_pass_loc), chunk_size):
            idx = filter_pass_loc[i:i + chunk_size]

            n_counts = adata[idx].obs[ncounts].values[:, None]
            X_view = adata[idx, coding_miRNA_loc].X
            X_norm = (X_view / n_counts * target_sum / norm_factor_vector)
            X_norm = sp.csr_matrix(X_norm)

            tokenized_cells.append([
                self.rank_genes(X_norm[i].data,
                                coding_miRNA_tokens[X_norm[i].indices])
                for i in range(X_norm.shape[0])
            ])

            # add custom attributes for subview to dict
            if self.custom_attr_name_dict is not None:
                for k in file_cell_metadata.keys():
                    file_cell_metadata[k] += adata[idx].obs[k].tolist()
            else:
                file_cell_metadata = None

        return tokenized_cells, file_cell_metadata



from .load import (
    distribution_dataset_load,
    generation_paired_dataset_load,
    three_dim_dataset_load,
    interaction_dataset_load,
    multi_dataset_load,
    property_dataset_load,
    bi_distribution_dataset_load,
    oracle_load,
    receptor_load,
    bm_group_load,
    general_load,
)
from .split import (
    create_fold,
    create_fold_setting_cold,
    create_combination_split,
    create_fold_time,
    create_scaffold_split,
    create_group_split,
    create_combination_generation_split,
)
from .misc import (
    print_sys,
    install,
    fuzzy_search,
    save_dict,
    load_dict,
    to_submission_format,
)
from .label_name_list import dataset2target_lists
from .label import (
    NegSample,
    label_transform,
    convert_y_unit,
    convert_to_log,
    convert_back_log,
    binarize,
    label_dist,
)
from .retrieve import (
    get_label_map,
    get_reaction_type,
    retrieve_label_name_list,
    retrieve_dataset_names,
    retrieve_all_benchmarks,
    retrieve_benchmark_names,
)
from .query import uniprot2seq, cid2smiles



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")

from . import generation_dataset
from ..metadata import dataset_names
from ..utils import create_fold


class RetroSyn(generation_dataset.PairedDataLoader):
    """Data loader class accessing to retro-synthetic prediction task."""

    def __init__(
        self,
        name,
        path="./data",
        print_stats=False,
        input_name="product",
        output_name="reactant",
    ):
        """To create an data loader object for forward reaction prediction task. The goal is to predict
        the reaction products given a set of reactants

        Args:
            name (str): the name of the datset
            path (str, optional): the path to the saved data file.
            print_stats (bool, optional): whether to print the basic statistics
            input_name (str, optional): the name of the column containing input molecular data (product)
            output_name (str, optional): the name of the column containing output molecular data (reactant)
        """
        super().__init__(name, path, print_stats, input_name, output_name)

    def get_split(
        self,
        method="random",
        seed=42,
        frac=[0.7, 0.1, 0.2],
        include_reaction_type=False,
    ):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split
            include_reaction_type (bool): whether or not to include reaction type in the split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """
        df = self.get_data(format="df")

        if include_reaction_type:
            from ..utils import get_reaction_type

            try:
                rt = get_reaction_type(self.name)
                df["reaction_type"] = rt
            except:
                raise ValueError(
                    "Reaction Type Unavailable for " + str(self.name) +
                    "! Please turn include_reaction_type to be false!")

        if method == "random":
            return create_fold(df, seed, frac)
        else:
            raise AttributeError("Please use the correct split method")



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from ..utils import bm_group_load, print_sys, fuzzy_search
from ..utils import (
    create_fold,
    create_fold_setting_cold,
    create_combination_split,
    create_fold_time,
    create_scaffold_split,
    create_group_split,
)
from ..metadata import (
    get_task2category,
    bm_metric_names,
    benchmark_names,
    bm_split_names,
    docking_target_info,
)
from ..evaluator import Evaluator


class BenchmarkGroup:
    """Boilerplate of benchmark group class. It downloads, processes, and loads a set of benchmark classes along with their splits. It also provides evaluators and train/valid splitters."""

    def __init__(self, name, path="./data", file_format="csv"):
        """create a benchmark group class object

        Args:
                name (str): the name of the benchmark group class
                path (str, optional): the path to save/load the benchkmark group dataset
                file_format (str, optional): designated file format for each dataset in the benchmark group

        """
        self.name = bm_group_load(name, path)
        self.path = os.path.join(path, self.name)
        self.datasets = benchmark_names[self.name]
        self.dataset_names = []
        self.file_format = file_format

        for task, datasets in self.datasets.items():
            for dataset in datasets:
                self.dataset_names.append(dataset)

    def __iter__(self):
        """iterator implementation to iterate over all benchmarks in the benchmark group

        Returns:
            BenchmarkGroup: self
        """
        self.index = 0
        self.num_datasets = len(self.dataset_names)
        return self

    def __next__(self):
        """iterator implementation to define the next benchmark

        Returns:
            dict: a dictionary of key values in a benchmark, namely the train_val file, test file and benchmark name

        Raises:
            StopIteration: stop when exceed the number of benchmarks
        """
        if self.index < self.num_datasets:
            dataset = self.dataset_names[self.index]
            print_sys("--- " + dataset + " ---")

            data_path = os.path.join(self.path, dataset)
            if not os.path.exists(data_path):
                os.mkdir(data_path)
            if self.file_format == "csv":
                train = pd.read_csv(os.path.join(data_path, "train_val.csv"))
                test = pd.read_csv(os.path.join(data_path, "test.csv"))
            elif self.file_format == "pkl":
                train = pd.read_pickle(os.path.join(data_path, "train_val.pkl"))
                test = pd.read_pickle(os.path.join(data_path, "test.pkl"))
            self.index += 1
            return {"train_val": train, "test": test, "name": dataset}
        else:
            raise StopIteration

    def get_train_valid_split(self, seed, benchmark, split_type="default"):
        """obtain training and validation split given a split type from train_val file

        Args:
            seed (int): the random seed of the data split
            benchmark (str): name of the benchmark
            split_type (str, optional): name of the split

        Returns:
            pd.DataFrame: the training and validation files

        Raises:
            NotImplementedError: split method not implemented
        """
        print_sys("generating training, validation splits...")
        dataset = fuzzy_search(benchmark, self.dataset_names)
        data_path = os.path.join(self.path, dataset)
        if self.file_format == "csv":
            train_val = pd.read_csv(os.path.join(data_path, "train_val.csv"))
        elif self.file_format == "pkl":
            train_val = pd.read_pickle(os.path.join(data_path, "train_val.pkl"))

        if split_type == "default":
            split_method = bm_split_names[self.name][dataset]
        else:
            split_method = split_type

        frac = [0.875, 0.125, 0.0]

        if split_method == "scaffold":
            out = create_scaffold_split(train_val,
                                        seed,
                                        frac=frac,
                                        entity="Drug")
        elif split_method == "random":
            out = create_fold(train_val, seed, frac=frac)
        elif split_method == "combination":
            out = create_combination_split(train_val, seed, frac=frac)
        elif split_method == "group":
            out = create_group_split(train_val,
                                     seed,
                                     holdout_frac=0.2,
                                     group_column="Year")
        else:
            raise NotImplementedError
        return out["train"], out["valid"]

    def get(self, benchmark):
        """get individual benchmark

        Args:
            benchmark (str): benchmark name

        Returns:
            dict: a dictionary of train_val, test dataframes and normalized name of the benchmark
        """
        dataset = fuzzy_search(benchmark, self.dataset_names)
        data_path = os.path.join(self.path, dataset)
        if self.file_format == "csv":
            train = pd.read_csv(os.path.join(data_path, "train_val.csv"))
            test = pd.read_csv(os.path.join(data_path, "test.csv"))
        elif self.file_format == "pkl":
            train = pd.read_pickle(os.path.join(data_path, "train_val.pkl"))
            test = pd.read_pickle(os.path.join(data_path, "test.pkl"))

        return {"train_val": train, "test": test, "name": dataset}

    def evaluate(self, pred, testing=True, benchmark=None, save_dict=True):
        """automatic evaluation function

        Args:
            pred (dict): a dictionary of benchmark name as the key and prediction array as the value
            testing (bool, optional): evaluate using testing set mode or validation set mode
            benchmark (str, optional): name of the benchmark
            save_dict (bool, optional): whether or not to save the evaluation result

        Returns:
            dict: a dictionary with key the benchmark name and value a dictionary of metrics to metric value

        Raises:
            ValueError: benchmark name not found
        """
        if testing:
            # test set evaluation
            metric_dict = bm_metric_names[self.name]
            out = {}
            for data_name, pred_ in pred.items():
                data_name = fuzzy_search(data_name, self.dataset_names)
                data_path = os.path.join(self.path, data_name)
                if self.file_format == "csv":
                    test = pd.read_csv(os.path.join(data_path, "test.csv"))
                elif self.file_format == "pkl":
                    test = pd.read_pickle(os.path.join(data_path, "test.pkl"))
                y = test.Y.values
                evaluator = eval("Evaluator(name = '" + metric_dict[data_name] +
                                 "')")
                out[data_name] = {
                    metric_dict[data_name]: round(evaluator(y, pred_), 3)
                }

                # If reporting accuracy across target classes
                if "target_class" in test.columns:
                    test["pred"] = pred_
                    for c in test["target_class"].unique():
                        data_name_subset = data_name + "_" + c
                        test_subset = test[test["target_class"] == c]
                        y_subset = test_subset.Y.values
                        pred_subset = test_subset.pred.values

                        evaluator = eval("Evaluator(name = '" +
                                         metric_dict[data_name_subset] + "')")
                        out[data_name_subset] = {
                            metric_dict[data_name_subset]:
                                round(evaluator(y_subset, pred_subset), 3)
                        }
            return out
        else:
            # validation set evaluation
            if benchmark is None:
                raise ValueError(
                    "Please specify the benchmark name for us to retrieve the standard metric!"
                )
            data_name = fuzzy_search(benchmark, self.dataset_names)
            metric_dict = bm_metric_names[self.name]
            evaluator = eval("Evaluator(name = '" + metric_dict[data_name] +
                             "')")
            return {metric_dict[data_name]: round(evaluator(true, pred), 3)}

    def evaluate_many(self,
                      preds,
                      save_file_name=None,
                      results_individual=None):
        """
        This function returns the data in a format needed to submit to the Leaderboard

        Args:
            preds (list of dict): list of dictionary of predictions, each item is the input to the evaluate function.
            save_file_name (str, optional): file name to save the result
            results_individual (list of dictionary, optional): if you already have results generated for each run, simply input here so that this function won't call the evaluation function again

        Returns:
            dict: a dictionary where key is the benchmark name and value is another dictionary where the key is the metric name and value is a list [mean, std].
        """
        min_requirement = 5

        if len(preds) < min_requirement:
            return ValueError("Must have predictions from at least " +
                              str(min_requirement) +
                              " runs for leaderboard submission")
        if results_individual is None:
            individual_results = []
            for pred in preds:
                retval = self.evaluate(pred)
                individual_results.append(retval)
        else:
            individual_results = results_individual

        given_dataset_names = list(individual_results[0].keys())
        aggregated_results = {}
        for dataset_name in given_dataset_names:
            my_results = []
            for individual_result in individual_results:
                my_result = list(individual_result[dataset_name].values())[0]
                my_results.append(my_result)
            u = np.mean(my_results)
            std = np.std(my_results)
            aggregated_results[dataset_name] = [round(u, 3), round(std, 3)]
        return aggregated_results



import numpy as np
import os

try:
    from rdkit import Chem, DataStructs
    from rdkit import rdBase

    rdBase.DisableLog("rdApp.error")
except:
    raise ImportError(
        "Please install rdkit by 'conda install -c conda-forge rdkit'! ")

try:
    import networkx as nx
except:
    raise ImportError("Please install networkx by 'pip install networkx'! ")

from ...utils import print_sys, install
from ._smartsPatts import smartsPatts

PubChemKeys = None


def InitKeys(keyList, keyDict):
    assert len(keyList) == len(keyDict.keys()), "length mismatch"
    for key in keyDict.keys():
        patt, count = keyDict[key]
        if patt != "?":
            sma = Chem.MolFromSmarts(patt)
            if not sma:
                print_sys("SMARTS parser error for key #%d: %s" % (key, patt))
            else:
                keyList[key - 1] = sma, count


def calcPubChemFingerPart1(mol, **kwargs):
    global PubChemKeys
    if PubChemKeys is None:
        PubChemKeys = [(None, 0)] * len(smartsPatts.keys())

        InitKeys(PubChemKeys, smartsPatts)
    ctor = kwargs.get("ctor", DataStructs.SparseBitVect)

    res = ctor(len(PubChemKeys) + 1)
    for i, (patt, count) in enumerate(PubChemKeys):
        if patt is not None:
            if count == 0:
                res[i + 1] = mol.HasSubstructMatch(patt)
            else:
                matches = mol.GetSubstructMatches(patt)
                if len(matches) > count:
                    res[i + 1] = 1
    return res


def func_1(mol, bits):
    ringSize = []
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    AllRingsAtom = mol.GetRingInfo().AtomRings()
    for ring in AllRingsAtom:
        ringSize.append(len(ring))
        for k, v in temp.items():
            if len(ring) == k:
                temp[k] += 1
    if temp[3] >= 2:
        bits[0] = 1
        bits[7] = 1
    elif temp[3] == 1:
        bits[0] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[14] = 1
        bits[21] = 1
    elif temp[4] == 1:
        bits[14] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[28] = 1
        bits[35] = 1
        bits[42] = 1
        bits[49] = 1
        bits[56] = 1
    elif temp[5] == 4:
        bits[28] = 1
        bits[35] = 1
        bits[42] = 1
        bits[49] = 1
    elif temp[5] == 3:
        bits[28] = 1
        bits[35] = 1
        bits[42] = 1
    elif temp[5] == 2:
        bits[28] = 1
        bits[35] = 1
    elif temp[5] == 1:
        bits[28] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[63] = 1
        bits[70] = 1
        bits[77] = 1
        bits[84] = 1
        bits[91] = 1
    elif temp[6] == 4:
        bits[63] = 1
        bits[70] = 1
        bits[77] = 1
        bits[84] = 1
    elif temp[6] == 3:
        bits[63] = 1
        bits[70] = 1
        bits[77] = 1
    elif temp[6] == 2:
        bits[63] = 1
        bits[70] = 1
    elif temp[6] == 1:
        bits[63] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[98] = 1
        bits[105] = 1
    elif temp[7] == 1:
        bits[98] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[112] = 1
        bits[119] = 1
    elif temp[8] == 1:
        bits[112] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[126] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[133] = 1
    else:
        pass

    return ringSize, bits


def func_2(mol, bits):
    """*Internal Use Only*
    saturated or aromatic carbon-only ring
    """
    AllRingsBond = mol.GetRingInfo().BondRings()
    ringSize = []
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        ######### saturated
        nonsingle = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                nonsingle = True
                break
        if nonsingle == False:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
        ######## aromatic carbon-only
        aromatic = True
        AllCarb = True
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "AROMATIC":
                aromatic = False
                break
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() != 6 or EndAtom.GetAtomicNum() != 6:
                AllCarb = False
                break
        if aromatic == True and AllCarb == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[1] = 1
        bits[8] = 1
    elif temp[3] == 1:
        bits[1] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[15] = 1
        bits[22] = 1
    elif temp[4] == 1:
        bits[15] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[29] = 1
        bits[36] = 1
        bits[43] = 1
        bits[50] = 1
        bits[57] = 1
    elif temp[5] == 4:
        bits[29] = 1
        bits[36] = 1
        bits[43] = 1
        bits[50] = 1
    elif temp[5] == 3:
        bits[29] = 1
        bits[36] = 1
        bits[43] = 1
    elif temp[5] == 2:
        bits[29] = 1
        bits[36] = 1
    elif temp[5] == 1:
        bits[29] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[64] = 1
        bits[71] = 1
        bits[78] = 1
        bits[85] = 1
        bits[92] = 1
    elif temp[6] == 4:
        bits[64] = 1
        bits[71] = 1
        bits[78] = 1
        bits[85] = 1
    elif temp[6] == 3:
        bits[64] = 1
        bits[71] = 1
        bits[78] = 1
    elif temp[6] == 2:
        bits[64] = 1
        bits[71] = 1
    elif temp[6] == 1:
        bits[64] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[99] = 1
        bits[106] = 1
    elif temp[7] == 1:
        bits[99] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[113] = 1
        bits[120] = 1
    elif temp[8] == 1:
        bits[113] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[127] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[134] = 1
    else:
        pass
    return ringSize, bits


def func_3(mol, bits):
    AllRingsBond = mol.GetRingInfo().BondRings()
    ringSize = []
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        ######### saturated
        nonsingle = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                nonsingle = True
                break
        if nonsingle == False:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
        ######## aromatic nitrogen-containing
        aromatic = True
        ContainNitro = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "AROMATIC":
                aromatic = False
                break
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() == 7 or EndAtom.GetAtomicNum() == 7:
                ContainNitro = True
                break
        if aromatic == True and ContainNitro == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[2] = 1
        bits[9] = 1
    elif temp[3] == 1:
        bits[2] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[16] = 1
        bits[23] = 1
    elif temp[4] == 1:
        bits[16] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[30] = 1
        bits[37] = 1
        bits[44] = 1
        bits[51] = 1
        bits[58] = 1
    elif temp[5] == 4:
        bits[30] = 1
        bits[37] = 1
        bits[44] = 1
        bits[51] = 1
    elif temp[5] == 3:
        bits[30] = 1
        bits[37] = 1
        bits[44] = 1
    elif temp[5] == 2:
        bits[30] = 1
        bits[37] = 1
    elif temp[5] == 1:
        bits[30] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[65] = 1
        bits[72] = 1
        bits[79] = 1
        bits[86] = 1
        bits[93] = 1
    elif temp[6] == 4:
        bits[65] = 1
        bits[72] = 1
        bits[79] = 1
        bits[86] = 1
    elif temp[6] == 3:
        bits[65] = 1
        bits[72] = 1
        bits[79] = 1
    elif temp[6] == 2:
        bits[65] = 1
        bits[72] = 1
    elif temp[6] == 1:
        bits[65] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[100] = 1
        bits[107] = 1
    elif temp[7] == 1:
        bits[100] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[114] = 1
        bits[121] = 1
    elif temp[8] == 1:
        bits[114] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[128] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[135] = 1
    else:
        pass
    return ringSize, bits


def func_4(mol, bits):
    AllRingsBond = mol.GetRingInfo().BondRings()
    ringSize = []
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        ######### saturated
        nonsingle = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                nonsingle = True
                break
        if nonsingle == False:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
        ######## aromatic heteroatom-containing
        aromatic = True
        heteroatom = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "AROMATIC":
                aromatic = False
                break
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() not in [
                    1, 6
            ] or EndAtom.GetAtomicNum() not in [
                    1,
                    6,
            ]:
                heteroatom = True
                break
        if aromatic == True and heteroatom == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[3] = 1
        bits[10] = 1
    elif temp[3] == 1:
        bits[3] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[17] = 1
        bits[24] = 1
    elif temp[4] == 1:
        bits[17] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[31] = 1
        bits[38] = 1
        bits[45] = 1
        bits[52] = 1
        bits[59] = 1
    elif temp[5] == 4:
        bits[31] = 1
        bits[38] = 1
        bits[45] = 1
        bits[52] = 1
    elif temp[5] == 3:
        bits[31] = 1
        bits[38] = 1
        bits[45] = 1
    elif temp[5] == 2:
        bits[31] = 1
        bits[38] = 1
    elif temp[5] == 1:
        bits[31] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[66] = 1
        bits[73] = 1
        bits[80] = 1
        bits[87] = 1
        bits[94] = 1
    elif temp[6] == 4:
        bits[66] = 1
        bits[73] = 1
        bits[80] = 1
        bits[87] = 1
    elif temp[6] == 3:
        bits[66] = 1
        bits[73] = 1
        bits[80] = 1
    elif temp[6] == 2:
        bits[66] = 1
        bits[73] = 1
    elif temp[6] == 1:
        bits[66] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[101] = 1
        bits[108] = 1
    elif temp[7] == 1:
        bits[101] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[115] = 1
        bits[122] = 1
    elif temp[8] == 1:
        bits[115] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[129] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[136] = 1
    else:
        pass
    return ringSize, bits


def func_5(mol, bits):
    ringSize = []
    AllRingsBond = mol.GetRingInfo().BondRings()
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        unsaturated = False
        nonaromatic = True
        Allcarb = True
        ######### unsaturated
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                unsaturated = True
                break
        ######## non-aromatic
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name == "AROMATIC":
                nonaromatic = False
                break
        ######## allcarb
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() != 6 or EndAtom.GetAtomicNum() != 6:
                Allcarb = False
                break
        if unsaturated == True and nonaromatic == True and Allcarb == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[4] = 1
        bits[11] = 1
    elif temp[3] == 1:
        bits[4] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[18] = 1
        bits[25] = 1
    elif temp[4] == 1:
        bits[18] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[32] = 1
        bits[39] = 1
        bits[46] = 1
        bits[53] = 1
        bits[60] = 1
    elif temp[5] == 4:
        bits[32] = 1
        bits[39] = 1
        bits[46] = 1
        bits[53] = 1
    elif temp[5] == 3:
        bits[32] = 1
        bits[39] = 1
        bits[46] = 1
    elif temp[5] == 2:
        bits[32] = 1
        bits[39] = 1
    elif temp[5] == 1:
        bits[32] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[67] = 1
        bits[74] = 1
        bits[81] = 1
        bits[88] = 1
        bits[95] = 1
    elif temp[6] == 4:
        bits[67] = 1
        bits[74] = 1
        bits[81] = 1
        bits[88] = 1
    elif temp[6] == 3:
        bits[67] = 1
        bits[74] = 1
        bits[81] = 1
    elif temp[6] == 2:
        bits[67] = 1
        bits[74] = 1
    elif temp[6] == 1:
        bits[67] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[102] = 1
        bits[109] = 1
    elif temp[7] == 1:
        bits[102] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[116] = 1
        bits[123] = 1
    elif temp[8] == 1:
        bits[116] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[130] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[137] = 1
    else:
        pass
    return ringSize, bits


def func_6(mol, bits):
    ringSize = []
    AllRingsBond = mol.GetRingInfo().BondRings()
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        unsaturated = False
        nonaromatic = True
        ContainNitro = False
        ######### unsaturated
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                unsaturated = True
                break
        ######## non-aromatic
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name == "AROMATIC":
                nonaromatic = False
                break
        ######## nitrogen-containing
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() == 7 or EndAtom.GetAtomicNum() == 7:
                ContainNitro = True
                break
        if unsaturated == True and nonaromatic == True and ContainNitro == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[5] = 1
        bits[12] = 1
    elif temp[3] == 1:
        bits[5] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[19] = 1
        bits[26] = 1
    elif temp[4] == 1:
        bits[19] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[33] = 1
        bits[40] = 1
        bits[47] = 1
        bits[54] = 1
        bits[61] = 1
    elif temp[5] == 4:
        bits[33] = 1
        bits[40] = 1
        bits[47] = 1
        bits[54] = 1
    elif temp[5] == 3:
        bits[33] = 1
        bits[40] = 1
        bits[47] = 1
    elif temp[5] == 2:
        bits[33] = 1
        bits[40] = 1
    elif temp[5] == 1:
        bits[33] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[68] = 1
        bits[75] = 1
        bits[82] = 1
        bits[89] = 1
        bits[96] = 1
    elif temp[6] == 4:
        bits[68] = 1
        bits[75] = 1
        bits[82] = 1
        bits[89] = 1
    elif temp[6] == 3:
        bits[68] = 1
        bits[75] = 1
        bits[82] = 1
    elif temp[6] == 2:
        bits[68] = 1
        bits[75] = 1
    elif temp[6] == 1:
        bits[68] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[103] = 1
        bits[110] = 1
    elif temp[7] == 1:
        bits[103] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[117] = 1
        bits[124] = 1
    elif temp[8] == 1:
        bits[117] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[131] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[138] = 1
    else:
        pass
    return ringSize, bits


def func_7(mol, bits):

    ringSize = []
    AllRingsBond = mol.GetRingInfo().BondRings()
    temp = {3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0}
    for ring in AllRingsBond:
        unsaturated = False
        nonaromatic = True
        heteroatom = False
        ######### unsaturated
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "SINGLE":
                unsaturated = True
                break
        ######## non-aromatic
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name == "AROMATIC":
                nonaromatic = False
                break
        ######## heteroatom-containing
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() not in [
                    1, 6
            ] or EndAtom.GetAtomicNum() not in [
                    1,
                    6,
            ]:
                heteroatom = True
                break
        if unsaturated == True and nonaromatic == True and heteroatom == True:
            ringSize.append(len(ring))
            for k, v in temp.items():
                if len(ring) == k:
                    temp[k] += 1
    if temp[3] >= 2:
        bits[6] = 1
        bits[13] = 1
    elif temp[3] == 1:
        bits[6] = 1
    else:
        pass
    if temp[4] >= 2:
        bits[20] = 1
        bits[27] = 1
    elif temp[4] == 1:
        bits[20] = 1
    else:
        pass
    if temp[5] >= 5:
        bits[34] = 1
        bits[41] = 1
        bits[48] = 1
        bits[55] = 1
        bits[62] = 1
    elif temp[5] == 4:
        bits[34] = 1
        bits[41] = 1
        bits[48] = 1
        bits[55] = 1
    elif temp[5] == 3:
        bits[34] = 1
        bits[41] = 1
        bits[48] = 1
    elif temp[5] == 2:
        bits[34] = 1
        bits[41] = 1
    elif temp[5] == 1:
        bits[34] = 1
    else:
        pass
    if temp[6] >= 5:
        bits[69] = 1
        bits[76] = 1
        bits[83] = 1
        bits[90] = 1
        bits[97] = 1
    elif temp[6] == 4:
        bits[69] = 1
        bits[76] = 1
        bits[83] = 1
        bits[90] = 1
    elif temp[6] == 3:
        bits[69] = 1
        bits[76] = 1
        bits[83] = 1
    elif temp[6] == 2:
        bits[69] = 1
        bits[76] = 1
    elif temp[6] == 1:
        bits[69] = 1
    else:
        pass
    if temp[7] >= 2:
        bits[104] = 1
        bits[111] = 1
    elif temp[7] == 1:
        bits[104] = 1
    else:
        pass
    if temp[8] >= 2:
        bits[118] = 1
        bits[125] = 1
    elif temp[8] == 1:
        bits[118] = 1
    else:
        pass
    if temp[9] >= 1:
        bits[132] = 1
    else:
        pass
    if temp[10] >= 1:
        bits[139] = 1
    else:
        pass
    return ringSize, bits


def func_8(mol, bits):

    AllRingsBond = mol.GetRingInfo().BondRings()
    temp = {"aromatic": 0, "heteroatom": 0}
    for ring in AllRingsBond:
        aromatic = True
        heteroatom = False
        for bondIdx in ring:
            if mol.GetBondWithIdx(bondIdx).GetBondType().name != "AROMATIC":
                aromatic = False
                break
        if aromatic == True:
            temp["aromatic"] += 1
        for bondIdx in ring:
            BeginAtom = mol.GetBondWithIdx(bondIdx).GetBeginAtom()
            EndAtom = mol.GetBondWithIdx(bondIdx).GetEndAtom()
            if BeginAtom.GetAtomicNum() not in [
                    1, 6
            ] or EndAtom.GetAtomicNum() not in [
                    1,
                    6,
            ]:
                heteroatom = True
                break
        if heteroatom == True:
            temp["heteroatom"] += 1
    if temp["aromatic"] >= 4:
        bits[140] = 1
        bits[142] = 1
        bits[144] = 1
        bits[146] = 1
    elif temp["aromatic"] == 3:
        bits[140] = 1
        bits[142] = 1
        bits[144] = 1
    elif temp["aromatic"] == 2:
        bits[140] = 1
        bits[142] = 1
    elif temp["aromatic"] == 1:
        bits[140] = 1
    else:
        pass
    if temp["aromatic"] >= 4 and temp["heteroatom"] >= 4:
        bits[141] = 1
        bits[143] = 1
        bits[145] = 1
        bits[147] = 1
    elif temp["aromatic"] == 3 and temp["heteroatom"] == 3:
        bits[141] = 1
        bits[143] = 1
        bits[145] = 1
    elif temp["aromatic"] == 2 and temp["heteroatom"] == 2:
        bits[141] = 1
        bits[143] = 1
    elif temp["aromatic"] == 1 and temp["heteroatom"] == 1:
        bits[141] = 1
    else:
        pass
    return bits


def calcPubChemFingerPart2(mol):  # 116-263

    bits = [0] * 148
    bits = func_1(mol, bits)[1]
    bits = func_2(mol, bits)[1]
    bits = func_3(mol, bits)[1]
    bits = func_4(mol, bits)[1]
    bits = func_5(mol, bits)[1]
    bits = func_6(mol, bits)[1]
    bits = func_7(mol, bits)[1]
    bits = func_8(mol, bits)

    return bits


def calcPubChemFingerAll(s):
    mol = Chem.MolFromSmiles(s)
    AllBits = [0] * 881
    res1 = list(calcPubChemFingerPart1(mol).ToBitString())
    for index, item in enumerate(res1[1:116]):
        if item == "1":
            AllBits[index] = 1
    for index2, item2 in enumerate(res1[116:734]):
        if item2 == "1":
            AllBits[index2 + 115 + 148] = 1
    res2 = calcPubChemFingerPart2(mol)
    for index3, item3 in enumerate(res2):
        if item3 == 1:
            AllBits[index3 + 115] = 1
    return np.array(AllBits)


def canonicalize(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is not None:
        return Chem.MolToSmiles(mol, isomericSmiles=True)
    else:
        return None


def smiles2pubchem(s):
    s = canonicalize(s)
    try:
        features = calcPubChemFingerAll(s)
    except:
        print("pubchem fingerprint not working for smiles: " + s +
              " convert to 0 vectors")
        features = np.zeros((881,))
    return np.array(features)



from transformers import PretrainedConfig, PreTrainedModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from typing import Optional, Dict


class ScGPTConfig(PretrainedConfig):
    model_type = "scgpt"

    def __init__(
            self,
            vocab_size=60697,
            embsize=512,
            d_hid=512,
            nlayers=12,
            nhead=8,
            max_seq_len=1536,
            dropout=0.0,
            pad_token_id=0,
            use_fast_transformer=True,
            input_emb_style="continuous",
            cell_emb_style="cls",  # output embedding vector with 
            norm_scheme="post",
            explicit_zero_prob=False,
            use_flash_attention=True,
            **kwargs):
        self.vocab_size = vocab_size
        self.embsize = embsize
        self.d_hid = d_hid
        self.nlayers = nlayers
        self.nhead = nhead
        self.max_seq_len = max_seq_len
        self.dropout = dropout
        self.pad_token_id = pad_token_id
        self.use_fast_transformer = use_fast_transformer
        if input_emb_style not in ["continuous"]:
            raise ValueError(
                f"Invalid input_emb_style: {input_emb_style}. Only continuous embeddings currently supported."
            )
        self.input_emb_style = input_emb_style
        self.cell_emb_style = cell_emb_style
        self.norm_scheme = norm_scheme
        self.explicit_zero_prob = explicit_zero_prob
        self.use_flash_attention = self.use_fast_transformer and torch.cuda.is_available(
        ) and use_flash_attention
        super().__init__(pad_token_id=pad_token_id, **kwargs)


class ExprDecoder(nn.Module):

    def __init__(self, d_model: int, explicit_zero_prob: bool = False):
        super().__init__()
        self.fc = nn.Sequential(
            nn.Linear(d_model, d_model),  # we don't use batch labels
            nn.LeakyReLU(),
            nn.Linear(d_model, d_model),
            nn.LeakyReLU(),
            nn.Linear(d_model, 1),
        )
        self.explicit_zero_prob = explicit_zero_prob
        if explicit_zero_prob:
            self.zero_logit = nn.Sequential(
                nn.Linear(d_model, d_model),
                nn.LeakyReLU(),
                nn.Linear(d_model, d_model),
                nn.LeakyReLU(),
                nn.Linear(d_model, 1),
            )

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        pred_value = self.fc(x).squeeze(-1)
        if not self.explicit_zero_prob:
            return {"pred": pred_value}
        zero_logits = self.zero_logit(x).squeeze(-1)
        zero_probs = torch.sigmoid(zero_logits)
        return {
            "pred": pred_value,
            "zero_probs": zero_probs
        }  # TODO: what about inference / bernoulli?


class FlashTransformerEncoderLayer(nn.Module):

    def __init__(self,
                 d_model,
                 nhead,
                 dim_feedforward,
                 dropout,
                 norm_scheme="post"):
        super().__init__()
        from flash_attn.flash_attention import FlashMHA

        self.self_attn = FlashMHA(
            embed_dim=d_model,
            num_heads=nhead,
            dropout=dropout,
            attention_dropout=dropout,
        )
        self.feed_forward = nn.Sequential(nn.Linear(d_model, dim_feedforward),
                                          nn.GELU(), nn.Dropout(dropout),
                                          nn.Linear(dim_feedforward, d_model))
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)
        self.norm_scheme = norm_scheme


# Helper class to ensure we have the correct attention structure
class MultiheadAttentionWithBias(nn.Module):

    def __init__(self, embed_dim, num_heads, dropout=0.0, batch_first=True):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first

        # Combined input projections for Q, K, V
        self.in_proj_weight = nn.Parameter(
            torch.empty((3 * embed_dim, embed_dim)))
        self.in_proj_bias = nn.Parameter(torch.empty(3 * embed_dim))

        # Output projection
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=True)

        self._reset_parameters()

    def _reset_parameters(self):
        # Initialize parameters following PyTorch's MultiheadAttention initialization
        nn.init.xavier_uniform_(self.in_proj_weight)
        nn.init.xavier_uniform_(self.out_proj.weight)
        nn.init.constant_(self.in_proj_bias, 0.)
        nn.init.constant_(self.out_proj.bias, 0.)

    def forward(self, query, key, value, key_padding_mask=None):
        return nn.functional.multi_head_attention_forward(
            query,
            key,
            value,
            self.embed_dim,
            self.num_heads,
            self.in_proj_weight,
            self.in_proj_bias,
            None,
            None,
            None,  # No bias_k, bias_v, or add_zero_attn
            self.dropout,
            self.out_proj.weight,
            self.out_proj.bias,
            training=self.training,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            batch_first=self.batch_first)[0]


from transformers import PreTrainedModel
import torch.nn as nn


class ScGPTPreTrainedModel(PreTrainedModel):
    config_class = ScGPTConfig
    base_model_prefix = "scgpt"

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)


class ScGPTModel(ScGPTPreTrainedModel):

    def __init__(self, config):
        super().__init__(config)

        # Gene name embeddings remain the same
        self.gene_encoder = nn.ModuleDict({
            "embedding":
                nn.Embedding(config.vocab_size,
                             config.embsize,
                             padding_idx=config.pad_token_id),
            "enc_norm":
                nn.LayerNorm(config.embsize)
        })

        # Value encoder remains the same
        if config.input_emb_style == "continuous":
            self.value_encoder = nn.ModuleDict({
                "linear1": nn.Linear(1, config.embsize),
                "linear2": nn.Linear(config.embsize, config.embsize),
                "norm": nn.LayerNorm(config.embsize),
                "dropout": nn.Dropout(config.dropout)
            })
        elif config.input_emb_style == "scaling":
            self.value_encoder = nn.Identity()
            raise Exception(
                "scaling input embedding style not supported because this model was trained on continuous style"
            )
        else:
            raise Exception("unsupported embedding style")

        # Modified transformer layers to use combined QKV projections
        # self.transformer = nn.ModuleDict({
        #     "layers": nn.ModuleList([
        #         nn.ModuleDict({
        #             "self_attn": MultiheadAttentionWithBias(
        #                 config.embsize,
        #                 config.nhead,
        #                 dropout=config.dropout,
        #                 batch_first=True
        #             ),
        #             "linear1": nn.Linear(config.embsize, config.d_hid),
        #             "linear2": nn.Linear(config.d_hid, config.embsize),
        #             "norm1": nn.LayerNorm(config.embsize),
        #             "norm2": nn.LayerNorm(config.embsize),
        #         }) for _ in range(config.nlayers)
        #     ])
        # })

        from torch.nn import TransformerEncoder, TransformerEncoderLayer
        self.transformer = TransformerEncoder(
            TransformerEncoderLayer(
                d_model=config.embsize,
                nhead=config.nhead,
                dim_feedforward=config.d_hid,
                dropout=config.dropout,
                batch_first=True,  # just for replication
            ),
            num_layers=config.nlayers)

        # Decoder remains the same
        self.expr_decoder = ExprDecoder(config.embsize,
                                        config.explicit_zero_prob)

        # we ignore cls_decoder because we do not pursue classification task
        # we also ignore mvc and similarity because we ignore generative tasks

        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        output_cell_emb: bool = True,
    ) -> Dict[str, torch.Tensor]:
        """
        Args:
            input_ids: Tensor of gene indices, shape [batch_size, seq_len]
            values: Tensor of expression values, shape [batch_size, seq_len]
            attention_mask: Optional mask tensor, shape [batch_size, seq_len]
            output_cell_emb: Whether to output cell embeddings

        Returns:
            Dictionary containing:
                - 'pred': Predicted expression values
                - 'cell_emb': Cell embeddings (if output_cell_emb=True)
                - 'zero_probs': Zero probabilities (if config.explicit_zero_prob=True)
        """
        # Gene embeddings
        gene_emb = self.gene_encoder["embedding"](input_ids)
        gene_emb = self.gene_encoder["enc_norm"](gene_emb)

        # Value encoding
        if hasattr(self, 'value_encoder'):
            values = values.unsqueeze(-1)  # Add feature dimension
            value_emb = self.value_encoder["linear1"](values)
            if "activation" in self.value_encoder:
                value_emb = self.value_encoder["activation"](value_emb)
            value_emb = self.value_encoder["linear2"](value_emb)
            value_emb = self.value_encoder["norm"](value_emb)
            value_emb = self.value_encoder["dropout"](value_emb)

            if self.config.input_emb_style == "continuous":
                hidden_states = gene_emb + value_emb
            else:  # "scaling", currrently not supported
                hidden_states = gene_emb * value_emb
        else:
            hidden_states = gene_emb

        # Convert attention_mask for transformer
        # Flash attention expects mask of 0s for tokens to attend to and 1s for tokens to ignore
        # if self.use_flash_attention and attention_mask is not None:
        #     if attention_mask.dtype != torch.bool:
        #         attention_mask = attention_mask.bool()
        #     attention_mask = ~attention_mask # we assume user follows huggingface convention for the attention mask

        # # Apply transformer layers
        # if self.use_flash_attention:
        #     for layer in self.transformer:
        #         hidden_states = layer(
        #             hidden_states,
        #             src_key_padding_mask=attention_mask
        #         )
        # else:
        hidden_states = self.transformer(hidden_states,
                                         src_key_padding_mask=attention_mask)

        # Get cell embeddings if requested
        output_dict = {}
        if output_cell_emb:
            if self.config.cell_emb_style == "cls":
                cell_emb = hidden_states[:, 0]
            elif self.config.cell_emb_style == "avg-pool":
                cell_emb = hidden_states.mean(dim=1)
            else:  # w-pool
                # Weighted pooling using input values as weights
                weights = F.softmax(values, dim=1).unsqueeze(-1)
                cell_emb = (hidden_states * weights).sum(dim=1)
            output_dict['cell_emb'] = cell_emb

        # Decode expression values
        decoder_output = self.expr_decoder(hidden_states)
        output_dict.update(decoder_output)

        return output_dict

    def _init_weights(self, module):
        super()._init_weights(module)
        if isinstance(module, nn.Linear):
            # Additional initialization for linear layers
            if module.bias is not None:
                nn.init.constant_(module.bias, 0)



"""
Class for customizations when transforming anndata format objects to pandas.DataFrame or other formats
"""

import pandas as pd
import numpy as np
from .data_feature_generator import DataFeatureGenerator


class AnnDataToDataFrame(DataFeatureGenerator):

    @classmethod
    def anndata_to_df(cls, dataset=None, obs_cols=None):
        if dataset is None:
            raise ValueError("dataset must be specified")
        adata = dataset
        if not isinstance(adata.X, np.ndarray):
            adata.X = adata.X.todense()
        df_main = pd.DataFrame(adata.X if adata.X is not None else adata.X,
                               columns=adata.var_names,
                               index=adata.obs_names)
        dfobs = pd.DataFrame(adata.obs,
                             columns=adata.obs.keys(),
                             index=adata.obs.index)
        if obs_cols is None:
            return df_main
        elif obs_cols == "ALL":
            return df_main.merge(dfobs,
                                 left_index=True,
                                 right_index=True,
                                 how='left')
        elif isinstance(obs_cols, list):
            return df_main.merge(dfobs[obs_cols],
                                 left_index=True,
                                 right_index=True,
                                 how='left')
        else:
            raise ValueError("obs_cols must be a list of column names or 'ALL'")



from .antibodyaff import AntibodyAff
from .catalyst import Catalyst
from .ddi import DDI
from .drugres import DrugRes
from .drugsyn import DrugSyn
from .dti import DTI
from .gda import GDA
from .mti import MTI
from .peptidemhc import PeptideMHC
from .ppi import PPI
from .proteinpeptide import ProteinPeptide
from .test_multi_pred import TestMultiPred
from .tcr_epi import TCREpitopeBinding
from .trialoutcome import TrialOutcome
from .perturboutcome import PerturbOutcome



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from . import bi_pred_dataset
from ..metadata import dataset_names
from ..dataset_configs.config_map import ConfigMap


class ProteinPeptide(bi_pred_dataset.DataLoader):
    """Data loader class to load datasets in Protein-Peptide Binding Prediction task.
    More info: TODO 

    Task Description: Regression.
                      Given the amino acid sequence of peptide and (TODO: complete),
                      predict the binding affinity.

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False


    """

    def __init__(self, name, path="./data", label_name=None, print_stats=False):
        """Create Protein-Peptide Prediction dataloader object"""
        label_name = label_name if label_name is not None else "KD (nm)"  # TODO: this column should be parsed into float and upper/lower
        cfm = ConfigMap()
        config = cfm.get(name)
        super().__init__(
            name,
            path,
            label_name,
            print_stats,
            dataset_names=dataset_names["ProteinPeptide"],
            data_config=config(),
        )
        self.entity1_name = "Sequence"  # peptide sequence
        self.entity2_name = "Protein Target"  # protein target label
        self.two_types = True

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)



from huggingface_hub import create_repo
from huggingface_hub import HfApi, snapshot_download, hf_hub_download
import os

deeppurpose_repo = [
    'hERG_Karim-Morgan',
    'hERG_Karim-CNN',
    'hERG_Karim-AttentiveFP',
    'BBB_Martins-AttentiveFP',
    'BBB_Martins-Morgan',
    'BBB_Martins-CNN',
    'CYP3A4_Veith-Morgan',
    'CYP3A4_Veith-CNN',
    'CYP3A4_Veith-AttentiveFP',
]

model_hub = ["Geneformer", "scGPT"]


class tdc_hf_interface:
    '''
    Example use cases:
    # initialize an interface object with HF repo name
    tdc_hf_herg = tdc_hf_interface("hERG_Karim-Morgan")
    # upload folder/files to this repo
    tdc_hf_herg.upload('./Morgan_herg_karim_optimal')
    # load deeppurpose model from this repo
    dp_model = tdc_hf_herg.load_deeppurpose('./data')
    dp_model.predict(XXX)
    '''

    def __init__(self, repo_name):
        self.repo_id = "tdc/" + repo_name
        try:
            self.model_name = repo_name.split('-')[1]
        except:
            self.model_name = repo_name

    def upload(self, folder_path):
        create_repo(repo_id=self.repo_id)
        api = HfApi()
        api.upload_folder(folder_path=folder_path,
                          path_in_repo="model",
                          repo_id=self.repo_id,
                          repo_type="model")

    def file_download(self, save_path, filename):
        model_ckpt = hf_hub_download(repo_id=self.repo_id,
                                     filename=filename,
                                     cache_dir=save_path)

    def repo_download(self, save_path):
        snapshot_download(repo_id=self.repo_id, cache_dir=save_path)

    def load(self):
        if self.model_name not in model_hub:
            raise Exception("this model is not in the TDC model hub GH repo.")
        elif self.model_name == "Geneformer":
            from transformers import AutoModelForMaskedLM
            model = AutoModelForMaskedLM.from_pretrained("tdc/Geneformer")
            return model
        elif self.model_name == "scGPT":
            from transformers import AutoConfig, AutoModel
            from .models.scgpt import ScGPTModel, ScGPTConfig
            AutoConfig.register("scgpt", ScGPTConfig)
            AutoModel.register(ScGPTConfig, ScGPTModel)
            model = AutoModel.from_pretrained("tdc/scGPT")
            return model
        raise Exception("Not implemented yet!")

    def load_deeppurpose(self, save_path):
        if self.repo_id[4:] in deeppurpose_repo:
            save_path = save_path + '/' + self.repo_id[4:]
            if not os.path.exists(save_path):
                os.mkdir(save_path)
            self.file_download(save_path, "model/model.pt")
            self.file_download(save_path, "model/config.pkl")

            save_path = save_path + '/models--tdc--' + self.repo_id[
                4:] + '/blobs/'
            file_name1 = save_path + os.listdir(save_path)[0]
            file_name2 = save_path + os.listdir(save_path)[1]

            if os.path.getsize(file_name1) > os.path.getsize(file_name2):
                model_file, config_file = file_name1, file_name2
            else:
                config_file, model_file = file_name1, file_name2

            os.rename(model_file, save_path + 'model.pt')
            os.rename(config_file, save_path + 'config.pkl')
            try:
                from DeepPurpose import CompoundPred
            except:
                raise ValueError(
                    "Please install DeepPurpose package following https://github.com/kexinhuang12345/DeepPurpose#installation"
                )

            net = CompoundPred.model_pretrained(path_dir=save_path)
            return net
        else:
            raise ValueError("This repo does not host a DeepPurpose model!")

    def predict_deeppurpose(self, model, drugs):
        try:
            from DeepPurpose import utils
        except:
            raise ValueError(
                "Please install DeepPurpose package following https://github.com/kexinhuang12345/DeepPurpose#installation"
            )
        if self.model_name == 'AttentiveFP':
            self.model_name = 'DGL_' + self.model_name
        X_pred = utils.data_process(X_drug=drugs,
                                    y=[0] * len(drugs),
                                    drug_encoding=self.model_name,
                                    split_method='no_split')
        y_pred = model.predict(X_pred)[0]
        return y_pred



import numpy as np
from typing import List, Tuple

from ...utils.load import pd_load, download_wrapper


def tokenize_batch(
    data: np.ndarray,
    gene_ids: np.ndarray,
    return_pt: bool = True,
    append_cls: bool = True,
    include_zero_gene: bool = False,
    cls_id: str = "<cls>",
) -> List[Tuple]:
    """
    Tokenize a batch of data. Returns a list of tuple (gene_id, count).

    Args:
        data (array-like): A batch of data, with shape (batch_size, n_features).
            n_features equals the number of all genes.
        gene_ids (array-like): A batch of gene ids, with shape (n_features,).
        return_pt (bool): Whether to return torch tensors of gene_ids and counts,
            default to True.

    Returns:
        list: A list of tuple (gene_names, counts) of non zero gene expressions.
    """
    download_wrapper("scgpt_vocab", "./data", ["scgpt_vocab"])
    vocab_map = pd_load("scgpt_vocab", "./data")
    if data.shape[1] != len(gene_ids):
        raise ValueError(
            f"Number of features in data ({data.shape[1]}) does not match "
            f"number of gene_ids ({len(gene_ids)}).")

    tokenized_data = []
    for i in range(len(data)):
        row = data[i]
        if include_zero_gene:
            values = row
            genes = gene_ids
        else:
            idx = np.nonzero(row)[0]
            values = row[idx]
            genes = gene_ids[idx]
        if append_cls:
            genes = np.insert(genes, 0, cls_id)
            values = np.insert(values, 0, 0)
        if return_pt:
            import torch
            genes = torch.tensor([vocab_map.get(x, 0) for x in genes],
                                 dtype=torch.int64)
            values = torch.from_numpy(values).float()
        tokenized_data.append((genes, values))
    return tokenized_data


class scGPTTokenizer:

    def __init__(self):
        pass

    @classmethod
    def tokenize_cell_vectors(cls, data, gene_names):
        """
        Tokenizing single-cell gene expression vectors formatted as anndata types
        """
        return tokenize_batch(data, gene_names)



from ..utils import general_load
"""
Resource class for the Eve Bio (https://evebio.org/) Pharmone Map.
"""


class PharmoneMap(object):

    def __init__(self, path="./data"):
        self.path = path

    def get_data(self):
        return general_load('evebio_pharmone_v1_detailed_result_table',
                            self.path, '\t')  # Load the Pharmone Map data

    def get_obs_metadata(self):
        return general_load("evebio_pharmone_v1_observed_points_table",
                            self.path, "\t")

    def get_control_data(self):
        return general_load("evebio_pharmone_v1_control_table", self.path,
                            "\t")  # Load the control data

    def get_compound_data(self):
        return general_load("evebio_pharmone_v1_compound_table", self.path,
                            "\t")

    def get_target_data(self):
        return general_load("evebio_pharmone_v1_target_table", self.path, "\t")



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import bi_distribution_dataset_load, print_sys
from ..utils import create_combination_generation_split


class DataLoader(base_dataset.DataLoader):
    """A base dataset loader class.

    Attributes:
        dataset_names (str): name of the dataset.
        name (str): The name fo the dataset.
        path (str): the path to save the data file.
        smiles_lst (list): a list of smiles strings as training data for distribution learning.
    """

    def __init__(
        self,
        name,
        path,
        print_stats,
        return_pocket=False,
        threshold=15,
        remove_Hs=True,
        keep_het=False,
        allowed_atom_list=[
            "C", "N", "O", "S", "H", "B", "Br", "Cl", "P", "I", "F"
        ],
    ):
        """To create a base dataloader object that each generation task can inherit from.

        Args:
            name (str): the name of the dataset.
            path (str): the path to save the data file.
            print_stats (bool): whether to print the basic statistics of the dataset.
            column_name (str): The name of the column containing smiles strings.
        """
        from ..metadata import multiple_molecule_dataset_names

        protein, ligand = bi_distribution_dataset_load(
            name,
            path,
            multiple_molecule_dataset_names,
            return_pocket,
            threshold,
            remove_Hs,
            keep_het,
            allowed_atom_list,
        )

        self.ligand = ligand
        self.protein = protein

        ### including fuzzy-search
        self.name = name
        self.path = path
        self.dataset_names = multiple_molecule_dataset_names
        self.return_pocket = return_pocket
        self.remove_Hs = remove_Hs
        self.keep_het = keep_het
        self.allowed_atom_list = allowed_atom_list
        if print_stats:
            self.print_stats()
        print_sys("Done!")

    def print_stats(self):
        """Print the basic statistics of the dataset."""
        print(
            "There are " + str(len(self.smiles_lst)) + " molecules ",
            flush=True,
            file=sys.stderr,
        )

    def get_data(self, format="df"):
        """Return the data from the whole dataset.

        Args:
            format (str, optional): the desired format for molecular data.

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset/a distionary for information

        Raises:
            AttributeError: Use the correct format as input (df, dict)
        """
        if format == "dict":
            return {"protein": self.protein, "ligand": self.ligand}
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """
        data = self.get_data(format="dict")
        protein, ligand = data["protein"], data["ligand"]

        if method == "random":
            return create_combination_generation_split(protein, ligand, seed,
                                                       frac)
        else:
            raise AttributeError("Please use the correct split method")
